import re
from pathlib import Path

import click
from click import argument, option

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

from sumo.cli.bandstats import kpt_str
from sumo.electronic_structure.effective_mass import (
    fit_effective_mass,
    get_fitting_data,
)

from amset.plot.base import write_plot_data
from amset.tools.common import image_type, path_type, zero_weighted_type

kpaths = click.Choice(["pymatgen", "seekpath"], case_sensitive=False)
x_properties = click.Choice(["doping", "temperature"], case_sensitive=False)
rate_plot_type = click.Choice(
    ["rate", "lifetime", "v2tau", "v2taudfde"], case_sensitive=False
)

_symprec = 0.01  # redefine symprec to avoid loading constants from file
_dos_estep = 0.01  # redefine symprec to avoid loading constants from file
_interpolation_factor = 5


def _all_or_int(value):
    if value == "all":
        return None
    elif value is None:
        return 0
    else:
        return int(value)


def _parse_kpoints(kpoints):
    return [
        [list(map(float, kpt.split())) for kpt in kpts.split(",")]
        for kpts in kpoints.split("|")
    ]


def _parse_kpoint_labels(labels):
    return [path.split(",") for path in labels.split("|")]


@click.group()
def plot():
    """
    Plot AMSET results, including scattering rates and band structures
    """


@plot.command()
@argument("filename", type=path_type)
@option("-t", "--temperature", default=0, help="temperature index [default: 0]")
@option("-d", "--doping", default=0, help="doping index [default: 0]")
@option("-l", "--line-density", default=100.0, help="band structure line density")
@option("-p", "--prefix", help="output filename prefix")
@option("--emin", help="minimum energy limit", type=float)
@option("--emax", help="maximum energy limit", type=float)
@option(
    "--amin",
    default=5e-5,
    help="minimum spectral weight for normalizing linewidth in 1/meV",
)
@option(
    "--amax",
    default=1e-1,
    help="maximum spectral weight for normalizing linewidth in 1/meV",
)
@option("--cmap", default="viridis", help="matplotlib colormap to use")
@option("--no-colorbar", default=True, is_flag=True, help="don't add a colorbar")
@option("--symprec", type=float, default=_symprec, help="interpolation factor")
@option("--kpath", type=kpaths, help="k-point path type")
@option(
    "--kpoints",
    type=_parse_kpoints,
    metavar="K",
    help="manual k-points list [e.g. '0 0 0, 0.5 0 0']",
)
@option(
    "--labels",
    type=_parse_kpoint_labels,
    metavar="L",
    help=r"labels for manual kpoints [e.g. '\Gamma,X']",
)
@option("--interpolation-factor", default=1, help="BoltzTraP interpolation factor")
@option("--width", default=3.2, help="figure width [default: 6]")
@option("--height", default=3.2, help="figure height [default: 6]")
@option("--directory", type=path_type, help="file output directory")
@option("--format", "image_format", default="pdf", type=image_type, help="image format")
@option("--style", help="matplotlib style specification")
@option("--no-base-style", default=False, is_flag=True, help="don't apply base style")
def lineshape(filename, **kwargs):
    """
    Plot band structures with electron lineshape
    """
    from amset.plot.lineshape import LineshapePlotter

    plotter = LineshapePlotter(
        filename, print_log=True, interpolation_factor=kwargs["interpolation_factor"]
    )

    kpath = get_kpath(
        plotter.structure,
        mode=kwargs["kpath"],
        symprec=kwargs["symprec"],
        kpt_list=kwargs["kpoints"],
        labels=kwargs["labels"],
    )

    plt = plotter.get_plot(
        kwargs["doping"],
        kwargs["temperature"],
        line_density=kwargs["line_density"],
        emin=kwargs["emin"],
        emax=kwargs["emax"],
        amin=kwargs["amin"],
        amax=kwargs["amax"],
        cmap=kwargs["cmap"],
        colorbar=kwargs["no_colorbar"],
        width=kwargs["width"],
        height=kwargs["height"],
        style=kwargs["style"],
        no_base_style=kwargs["no_base_style"],
        kpath=kpath,
    )

    save_plot(
        plt, "lineshape", kwargs["directory"], kwargs["prefix"], kwargs["image_format"]
    )
    return plt


@plot.command()
@argument("filename", type=path_type)
@option("-l", "--line-density", default=100.0, help="band structure line density")
@option("--emin", default=-6.0, help="minimum energy limit")
@option("--emax", default=6.0, help="maximum energy limit")
@option("--symprec", type=float, default=_symprec, help="interpolation factor")
@option(
    "--print-log/--no-print-log",
    default=True,
    help="whether to print interpolation log",
)
@option("--kpath", type=kpaths, help="k-point path type")
@option(
    "--kpoints",
    type=_parse_kpoints,
    metavar="K",
    help="manual k-points list [e.g. '0 0 0, 0.5 0 0']",
)
@option(
    "--labels",
    type=_parse_kpoint_labels,
    metavar="L",
    help=r"labels for manual kpoints [e.g. '\Gamma,X']",
)
@option(
    "--interpolation-factor",
    default=_interpolation_factor,
    type=float,
    help="BoltzTraP interpolation factor",
)
@option("--energy-cutoff", type=float, help="interpolation energy cutoff in eV")
@option(
    "-z",
    "--zero-weighted-kpoints",
    help="how to process zero-weighted k-points",
    type=zero_weighted_type,
)
@option("--plot-dos", is_flag=True, help="whether to also plot the density of states")
@option(
    "--dos-kpoints",
    default="50",
    help="k-point length cutoff or mesh for density of states",
)
@option("--dos-estep", default=_dos_estep, help="dos energy step size")
@option("--dos-aspect", default=3.0, help="aspect ratio for the density of states")
@option(
    "--no-zero-to-efermi",
    "zero_to_efermi",
    is_flag=True,
    default=True,
    help="don't set the Fermi level to zero",
)
@option("--vbm-cbm-marker", is_flag=True, help="add a marker at the CBM and VBM")
@option("--stats", is_flag=True, help="print effective mass and band gap")
@option("--width", default=6.0, help="figure width [default: 6]")
@option("--height", default=6.0, help="figure height [default: 6]")
@option("-p", "--prefix", help="output filename prefix")
@option("--directory", type=path_type, help="file output directory")
@option("--format", "image_format", default="pdf", type=image_type, help="image format")
@option("--style", help="path to matplotlib style specification")
@option("--no-base-style", default=False, is_flag=True, help="don't apply base style")
def band(filename, **kwargs):
    """
    Plot interpolate band structure from vasprun file
    """
    from amset.constants import defaults
    from amset.plot.electronic_structure import ElectronicStructurePlotter

    zwk_mode = kwargs.pop("zero_weighted_kpoints")
    if not zwk_mode:
        zwk_mode = defaults["zero_weighted_kpoints"]

    plotter_kwargs = {
        "print_log": kwargs["print_log"],
        "interpolation_factor": kwargs["interpolation_factor"],
        "symprec": kwargs["symprec"],
        "energy_cutoff": kwargs["energy_cutoff"],
    }
    if is_vasprun_file(filename):
        plotter = ElectronicStructurePlotter.from_vasprun(
            filename, zero_weighted_kpoints=zwk_mode, **plotter_kwargs
        )
    elif is_band_structure_data_file(filename):
        plotter = ElectronicStructurePlotter.from_band_structure_data(
            filename, **plotter_kwargs
        )
    else:
        click.echo(
            "Unrecognised filetype, expecting a vasprun.xml or band_structure_data.json"
            " file."
        )
        raise click.Abort()

    kpath = get_kpath(
        plotter.structure,
        mode=kwargs["kpath"],
        symprec=kwargs["symprec"],
        kpt_list=kwargs["kpoints"],
        labels=kwargs["labels"],
    )

    dos_kpoints = _get_dos_kpoints(plotter.structure, kwargs["dos_kpoints"])

    plt, bs_plotter = plotter.get_plot(
        plot_band_structure=True,
        plot_dos=kwargs["plot_dos"],
        line_density=kwargs["line_density"],
        dos_kpoints=dos_kpoints,
        dos_estep=kwargs["dos_estep"],
        dos_aspect=kwargs["dos_aspect"],
        zero_to_efermi=kwargs["zero_to_efermi"],
        vbm_cbm_marker=kwargs["vbm_cbm_marker"],
        emin=kwargs["emin"],
        emax=kwargs["emax"],
        width=kwargs["width"],
        height=kwargs["height"],
        style=kwargs["style"],
        no_base_style=kwargs["no_base_style"],
        kpath=kpath,
        return_plotter=True,
    )

    save_plot(
        plt, "band", kwargs["directory"], kwargs["prefix"], kwargs["image_format"]
    )

    if kwargs["stats"]:
        _log_band_stats(bs_plotter._bs[0])

    return plt


@plot.command()
@argument("filename", type=path_type)
@option("-d", "--doping-idx", metavar="N", help="doping index to plot")
@option("-t", "--temperature-idx", metavar="N", help="temperature index to plot")
@option(
    "-s",
    "--separate-rates",
    is_flag=True,
    default=False,
    help="whether to separate scattering mechanisms",
)
@option(
    "--plot-type",
    type=rate_plot_type,
    default="rate",
    help="what to plot on the y-axis",
)
@option(
    "--total/--no-total", "plot_total_rate", default=False, help="plot the total rate"
)
@option(
    "--dfde/--no-dfde",
    "show_dfde",
    default=False,
    help="indicate the Fermi-Dirac derivative weight through color saturation",
)
@option("--use-symbol/--no-use-symbol", default=False, help="use symbols for labels")
@option(
    "--fd-tols/--no-fd-tols",
    "plot_fd_tols",
    default=False,
    help="plot the Fermi-Dirac tolerance limits",
)
@option(
    "--normalize/--no-normalize",
    "normalize_energy",
    default=True,
    help="normalize energies to the VBM/Fermi level",
)
@option("--pad", type=float, default=0.1, help="pad in % for y-axis limits")
@option("--ymin", default=None, type=float, help="minimum y-axis limit")
@option("--ymax", default=None, type=float, help="maximum y-axis limit")
@option("--xmin", default=None, type=float, help="minimum x-axis limit")
@option("--xmax", default=None, type=float, help="maximum x-axis limit")
@option("--total-color", help="color for total scattering rate")
@option("-p", "--prefix", help="output filename prefix")
@option("--directory", type=path_type, help="file output directory")
@option("--format", "image_format", default="pdf", type=image_type, help="image format")
@option("--style", help="path to matplotlib style specification")
@option("--no-base-style", default=False, is_flag=True, help="don't apply base style")
@option("--font", "fonts", help="font to use")
def rates(filename, **kwargs):
    """
    Plot scattering rates
    """
    from amset.plot.rates import RatesPlotter

    save_kwargs = [kwargs.pop(d) for d in ["directory", "prefix", "image_format"]]

    kwargs["doping_idx"] = _all_or_int(kwargs["doping_idx"])
    kwargs["temperature_idx"] = _all_or_int(kwargs["temperature_idx"])

    pad = kwargs.pop("pad")
    use_symbol = kwargs.pop("use_symbol")
    plotter = RatesPlotter(filename, pad=pad, use_symbol=use_symbol)
    plt = plotter.get_plot(**kwargs)
    plt.subplots_adjust(hspace=0.3, wspace=0.4)
    save_plot(plt, "rates", *save_kwargs)
    return plt


@plot.command()
@argument("filename", type=path_type)
@option(
    "-d", "--doping-idx", metavar="N", help="doping indices to plot (space separated)"
)
@option(
    "-t",
    "--temperature-idx",
    metavar="N",
    help="temperature indices to plot (space separated)",
)
@option("-x", "--x-property", type=x_properties, help="property to plot on x-axis")
@option("--grid", nargs=2, type=int, help="subplot grid (nrows, ncols)")
@option("--n-type", is_flag=True, default=False, help="plot n-type results only")
@option("--p-type", is_flag=True, default=False, help="plot p-type results only")
@option(
    "--average/--no-average",
    default=True,
    help="whether to average tensor transport properties",
)
@option("--pad", type=float, default=0.05, help="pad in % for axis limits")
@option("--use-symbol/--no-use-symbol", default=False, help="use symbols for labels")
@option("--xlabel", help="x-axis label")
@option("--xmin", type=float, help="minimum x-axis limit")
@option("--xmax", type=float, help="maximum x-axis limit")
@option("--logx/--no-log-x", default=None, help="log x-axis")
@option("--conductivity/--no-conductivity", default=True, help="plot conductivity")
@option("--seebeck/--no-seebeck", default=True, help="plot Seebeck coefficient")
@option(
    "--thermal-conductivity/--no-thermal-conductivity",
    default=True,
    help="plot electronic thermal conductivity",
)
@option("--mobility/--no-mobility", default=False, help="plot electron mobility")
@option("--power-factor/--no-power-factor", default=False, help="plot power factor")
@option("--conductivity-label", help="conductivity y-axis label")
@option("--conductivity-min", type=float, help="minimum conductivity y-axis limit")
@option("--conductivity-max", type=float, help="maximum conductivity y-axis limit")
@option(
    "--log-conductivity/--no-log-conductivity",
    default=None,
    help="plot log conductivity",
)
@option("--seebeck-label", help="Seebeck coefficient y-axis label")
@option("--seebeck-min", type=float, help="minimum Seebeck y-axis limit")
@option("--seebeck-max", type=float, help="maximum Seebeck y-axis limit")
@option("--log-seebeck/--no-log-seebeck", default=None, help="plot log Seebeck")
@option("--thermal-conductivity-label", help="thermal conductivity y-axis label")
@option(
    "--thermal-conductivity-min",
    type=float,
    help="minimum thermal conductivity y-axis limit",
)
@option(
    "--thermal-conductivity-max",
    type=float,
    help="maximum thermal conductivity y-axis limit",
)
@option(
    "--log-thermal-conductivity/--no-log-thermal-conductivity",
    default=None,
    help="plot log thermal conductivity",
)
@option("--mobility-label", help="mobility y-axis label")
@option("--mobility-min", type=float, help="minimum mobility y-axis limit")
@option("--mobility-max", type=float, help="maximum mobility y-axis limit")
@option("--log-mobility/--no-log-mobility", default=None, help="plot log mobility")
@option("--power-factor-label", help="power factor y-axis label")
@option("--power-factor-min", type=float, help="minimum power factor y-axis limit")
@option("--power-factor-max", type=float, help="maximum power factor y-axis limit")
@option(
    "--log-power-factor/--no-log-power-factor",
    default=None,
    help="plot log power factor",
)
@option("-p", "--prefix", help="output filename prefix")
@option("--width", type=float, help="figure width")
@option("--height", type=float, help="figure height")
@option("--directory", type=path_type, help="file output directory")
@option(
    "--gnuplot", default=False, is_flag=True, help="write plot data in gnuplot format"
)
@option("--format", "image_format", default="pdf", type=image_type, help="image format")
@option("--style", help="path to matplotlib style specification")
@option("--no-base-style", default=False, is_flag=True, help="don't apply base style")
@option("--font", "fonts", help="font to use")
def transport(filename, **kwargs):
    """
    Plot transport properties
    """
    from amset.plot.transport import TransportPlotter

    save_kwargs = {d: kwargs.pop(d) for d in ("directory", "prefix", "image_format")}

    properties = []
    for prop in (
        "conductivity",
        "seebeck",
        "thermal_conductivity",
        "mobility",
        "power_factor",
    ):
        if kwargs.pop(prop):
            properties.append(prop.replace("_", " "))

    kwargs["temperature_idx"] = _to_int(kwargs["temperature_idx"])
    kwargs["doping_idx"] = _to_int(kwargs["doping_idx"])

    gnuplot = kwargs.pop("gnuplot")
    pad = kwargs.pop("pad")
    use_symbol = kwargs.pop("use_symbol")
    average = kwargs.pop("average")
    doping_type = _get_doping_type(kwargs.pop("n_type"), kwargs.pop("p_type"))

    plotter = TransportPlotter(
        filename, pad=pad, use_symbol=use_symbol, average=average
    )
    plt, plot_data = plotter.get_plot(
        properties=properties, doping_type=doping_type, return_plot_data=True, **kwargs
    )
    plt.subplots_adjust(hspace=0.3, wspace=0.4)
    save_plot(plt, "transport", **save_kwargs)

    if gnuplot:
        write_plot_data(
            plot_data, prefix=save_kwargs["prefix"], directory=save_kwargs["directory"]
        )
        return plt, plot_data

    return plt


@plot.command()
@argument("filename", type=path_type)
@option(
    "-d", "--doping-idx", metavar="N", help="doping indices to plot (space separated)"
)
@option(
    "-t",
    "--temperature-idx",
    metavar="N",
    help="temperature indices to plot (space separated)",
)
@option("--n-type", is_flag=True, default=False, help="plot n-type results only")
@option("--p-type", is_flag=True, default=False, help="plot p-type results only")
@option("-x", "--x-property", type=x_properties, help="property to plot on x-axis")
@option(
    "--separate-mobility/--no-separate-mobility",
    default=True,
    help="whether to separate mobility for each scattering mechanisms",
)
@option("--title/--no-title", default=True, help="put a title above each subplot")
@option("--grid", nargs=2, type=int, help="subplot grid (nrows, ncols)")
@option(
    "--average/--no-average",
    default=True,
    help="whether to average tensor transport properties",
)
@option("--pad", type=float, default=0.05, help="pad in % for axis limits")
@option("--use-symbol/--no-use-symbol", default=False, help="use symbols for labels")
@option("--xlabel", help="x-axis label")
@option("--xmin", type=float, help="minimum x-axis limit")
@option("--xmax", type=float, help="maximum x-axis limit")
@option("--logx/--no-log-x", default=None, help="log x-axis")
@option("--ylabel", help="y-axis label")
@option("--ymin", type=float, help="minimum y-axis limit")
@option("--ymax", type=float, help="maximum y-axis limit")
@option("--total-color", help="color for total mobility line")
@option("--logy/--no-log-y", default=None, help="log y-axis")
@option("-p", "--prefix", help="output filename prefix")
@option("--width", type=float, help="figure width")
@option("--height", type=float, help="figure height")
@option("--directory", type=path_type, help="file output directory")
@option("--format", "image_format", default="pdf", type=image_type, help="image format")
@option("--style", help="path to matplotlib style specification")
@option("--no-base-style", default=False, is_flag=True, help="don't apply base style")
@option("--font", "fonts", help="font to use")
def mobility(filename, **kwargs):
    """
    Plot mobility in more detail
    """
    from amset.plot.mobility import MobilityPlotter

    save_kwargs = [kwargs.pop(d) for d in ("directory", "prefix", "image_format")]

    kwargs["temperature_idx"] = _to_int(kwargs["temperature_idx"])
    kwargs["doping_idx"] = _to_int(kwargs["doping_idx"])

    doping_type = _get_doping_type(kwargs.pop("n_type"), kwargs.pop("p_type"))

    init_kwargs = {
        d: kwargs.pop(d) for d in ["use_symbol", "average", "separate_mobility", "pad"]
    }
    plotter = MobilityPlotter(filename, **init_kwargs)
    plt = plotter.get_plot(doping_type=doping_type, **kwargs)
    plt.subplots_adjust(hspace=0.3, wspace=0.4)
    save_plot(plt, "mobility", *save_kwargs)
    return plt


@plot.command()
@argument("filenames", nargs=-1, type=path_type)
@option("-d", "--doping-idx", metavar="N", help="doping indices to plot")
@option("-t", "--temperature-idx", metavar="N", help="temperature indices to plot")
@option("-x", "--x-property", type=x_properties, help="property to plot on x-axis")
@option("--grid", nargs=2, type=int, help="subplot grid (nrows, ncols)")
@option("--pad", type=float, default=0.05, help="pad in % for axis limits")
@option("--labels", help="labels to use (space separated)")
@option("--use-symbol/--no-use-symbol", default=False, help="use symbols for labels")
@option("--xlabel", help="x-axis label")
@option("--xmin", type=float, help="minimum x-axis limit")
@option("--xmax", type=float, help="maximum x-axis limit")
@option("--logx/--no-log-x", default=None, help="log x-axis")
@option("--conductivity/--no-conductivity", default=True, help="plot conductivity")
@option("--seebeck/--no-seebeck", default=True, help="plot Seebeck coefficient")
@option(
    "--thermal-conductivity/--no-thermal-conductivity",
    default=True,
    help="plot electronic thermal conductivity",
)
@option("--mobility/--no-mobility", default=False, help="plot electron mobility")
@option("--power-factor/--no-power-factor", default=False, help="plot power factor")
@option("--conductivity-label", help="conductivity y-axis label")
@option("--conductivity-min", type=float, help="minimum conductivity y-axis limit")
@option("--conductivity-max", type=float, help="maximum conductivity y-axis limit")
@option(
    "--log-conductivity/--no-log-conductivity",
    default=None,
    help="plot log conductivity",
)
@option("--seebeck-label", help="Seebeck coefficient y-axis label")
@option("--seebeck-min", type=float, help="minimum Seebeck y-axis limit")
@option("--seebeck-max", type=float, help="maximum Seebeck y-axis limit")
@option("--log-seebeck/--no-log-seebeck", default=None, help="plot log Seebeck")
@option("--thermal-conductivity-label", help="thermal conductivity y-axis label")
@option(
    "--thermal-conductivity-min",
    type=float,
    help="minimum thermal conductivity y-axis limit",
)
@option(
    "--thermal-conductivity-max",
    type=float,
    help="maximum thermal conductivity y-axis limit",
)
@option(
    "--log-thermal-conductivity/--no-log-thermal-conductivity",
    default=None,
    help="plot log thermal conductivity",
)
@option("--mobility-label", help="mobility y-axis label")
@option("--mobility-min", type=float, help="minimum mobility y-axis limit")
@option("--mobility-max", type=float, help="maximum mobility y-axis limit")
@option("--log-mobility/--no-log-mobility", default=None, help="plot log mobility")
@option("--power-factor-label", help="power factor y-axis label")
@option("--power-factor-min", type=float, help="minimum power factor y-axis limit")
@option("--power-factor-max", type=float, help="maximum power factor y-axis limit")
@option(
    "--log-power-factor/--no-log-power-factor",
    default=None,
    help="plot log power factor",
)
@option("-p", "--prefix", help="output filename prefix")
@option("--width", type=float, help="figure width")
@option("--height", type=float, help="figure height")
@option("--directory", type=path_type, help="file output directory")
@option("--format", "image_format", default="pdf", type=image_type, help="image format")
@option("--style", help="path to matplotlib style specification")
@option("--no-base-style", default=False, is_flag=True, help="don't apply base style")
@option("--font", "fonts", help="font to use")
def convergence(filenames, **kwargs):
    """
    Plot transport properties
    """
    from amset.plot.convergence import ConvergencePlotter

    save_kwargs = [kwargs.pop(d) for d in ("directory", "prefix", "image_format")]

    properties = []
    for prop in (
        "conductivity",
        "seebeck",
        "thermal_conductivity",
        "mobility",
        "power_factor",
    ):
        if kwargs.pop(prop):
            properties.append(prop.replace("_", " "))

    if kwargs["labels"] is None:
        # try to determine labels from file names
        try:
            kwargs["labels"] = get_labels_from_filenames(filenames)
        except ValueError:
            click.echo("Could not determine labels from filenames. Use --labels option")

    kwargs["temperature_idx"] = _to_int(kwargs["temperature_idx"])
    kwargs["doping_idx"] = _to_int(kwargs["doping_idx"])

    pad = kwargs.pop("pad")
    use_symbol = kwargs.pop("use_symbol")
    plotter = ConvergencePlotter(filenames, pad=pad, use_symbol=use_symbol)
    plt = plotter.get_plot(properties=properties, **kwargs)
    plt.subplots_adjust(hspace=0.3, wspace=0.4)
    save_plot(plt, "convergence", *save_kwargs)
    return plt


def get_labels_from_filenames(filenames):
    labels = []
    for filename in filenames:
        match = re.search(r"\d+x\d+x\d+", str(filename))
        if not match:
            raise ValueError("Problem extracting mesh")
        labels.append(match[0])
    return labels


def _to_int(idxs):
    if idxs is not None:
        idxs = list(map(int, idxs.split()))
    return idxs


def save_plot(plt, name, directory, prefix, image_format):
    if prefix:
        prefix += "_"
    else:
        prefix = ""
    filename = Path(f"{prefix}{name}.{image_format}")

    if directory:
        filename = directory / filename

    plt.savefig(filename, format=image_format, bbox_inches="tight", dpi=400)


def is_vasprun_file(filename):
    return "xml" in filename


def is_band_structure_data_file(filename):
    return "band_structure_data" in filename


def get_kpath(structure, mode="pymatgen", symprec=_symprec, kpt_list=None, labels=None):
    r"""Get a Kpath object

    If a manual list of kpoints is supplied using the `kpt_list`
    variable, the `mode` option will be ignored.

    Args:
        structure: The structure.
        mode: Method used for calculating the
            high-symmetry path. The options are:

            pymatgen
                Use the paths from pymatgen.

            seekpath
                Use the paths from SeeK-path.

        symprec: The tolerance for determining the crystal symmetry.
        kpt_list: List of k-points to use, formatted as a list of subpaths, each
            containing a list of fractional k-points. For example:

            ```
                [ [[0., 0., 0.], [0., 0., 0.5]], [[0.5, 0., 0.], [0.5, 0.5, 0.]] ]
            ```

            will return points along `0 0 0 -> 0 0 1/2 | 1/2 0 0 -> 1/2 1/2 0`
        labels: The k-point labels. These should be provided as a list of strings for
            each subpath of the overall path. For example::

            ```
                [ ['Gamma', 'Z'], ['X', 'M'] ]
            ```

            combined with the above example for `kpt_list` would indicate the path:
            `Gamma -> Z | X -> M`. If no labels are provided, letters from A -> Z will
            be used instead.

    Returns:
        A Kpath object.
    """
    from sumo.symmetry import CustomKpath, PymatgenKpath, SeekpathKpath

    if kpt_list:
        kpath = CustomKpath(structure, kpt_list, labels, symprec=symprec)
    elif mode == "seekpath":
        kpath = SeekpathKpath(structure, symprec=symprec)
    else:
        kpath = PymatgenKpath(structure, symprec=symprec)

    return kpath


def _get_dos_kpoints(structure, kpoints):
    from amset.electronic_structure.kpoints import get_kpoint_mesh

    if kpoints is None:
        return None

    splits = kpoints.split()
    if len(splits) == 1:
        return get_kpoint_mesh(structure, float(splits[0]))
    else:
        return list(map(int, splits))


def _get_doping_type(n_type, p_type):
    if n_type and p_type:
        return None
    elif n_type:
        return "n"
    elif p_type:
        return "p"
    return None


def _log_band_stats(bs, parabolic=False, num_sample_points=3):
    if bs.is_metal():
        click.echo("\nSystem is metallic, cannot not print band stats")
        return
    else:
        click.echo("\nBand structure information\n~~~~~~~~~~~~~~~~~~~~~~~~~~\n")

    _log_band_gap_information(bs)

    vbm_data = bs.get_vbm()
    cbm_data = bs.get_cbm()

    click.echo("\nValence band maximum:")
    _log_band_edge_information(bs, vbm_data)

    click.echo("\nConduction band minimum:")
    _log_band_edge_information(bs, cbm_data)

    if parabolic:
        click.echo("\nUsing parabolic fitting of the band edges")
    else:
        click.echo("\nUsing nonparabolic fitting of the band edges")

    # Work out where the hole and electron band edges are.
    # Fortunately, pymatgen does this for us. Points at which to calculate
    # the effective mass are identified as a tuple of:
    # (spin, band_index, kpoint_index)
    hole_extrema = []
    for spin, bands in vbm_data["band_index"].items():
        hole_extrema.extend(
            [
                (spin, band_idx, kpoint)
                for band_idx in bands
                for kpoint in vbm_data["kpoint_index"]
            ]
        )

    elec_extrema = []
    for spin, bands in cbm_data["band_index"].items():
        elec_extrema.extend(
            [
                (spin, band_idx, kpoint)
                for band_idx in bands
                for kpoint in cbm_data["kpoint_index"]
            ]
        )

    # extract the data we need for fitting from the band structure
    hole_data = []
    for extrema in hole_extrema:
        hole_data.extend(
            get_fitting_data(bs, *extrema, num_sample_points=num_sample_points)
        )

    elec_data = []
    for extrema in elec_extrema:
        elec_data.extend(
            get_fitting_data(bs, *extrema, num_sample_points=num_sample_points)
        )

    # calculate the effective masses and log the information
    click.echo("\nHole effective masses:")
    for data in hole_data:
        eff_mass = fit_effective_mass(
            data["distances"], data["energies"], parabolic=parabolic
        )
        data["effective_mass"] = eff_mass
        _log_effective_mass_data(data, bs.is_spin_polarized, mass_type="m_h")

    click.echo("\nElectron effective masses:")
    for data in elec_data:
        eff_mass = fit_effective_mass(
            data["distances"], data["energies"], parabolic=parabolic
        )
        data["effective_mass"] = eff_mass
        _log_effective_mass_data(data, bs.is_spin_polarized)

    return {"hole_data": hole_data, "electron_data": elec_data}


def _log_band_gap_information(bs):
    from pymatgen.electronic_structure.core import Spin

    bg_data = bs.get_band_gap()
    if not bg_data["direct"]:
        click.echo("Indirect band gap: {:.3f} eV".format(bg_data["energy"]))

    direct_data = bs.get_direct_band_gap_dict()
    if bs.is_spin_polarized:
        direct_bg = min(spin_data["value"] for spin_data in direct_data.values())
        click.echo(f"Direct band gap: {direct_bg:.3f} eV")

        for spin, spin_data in direct_data.items():
            direct_kindex = spin_data["kpoint_index"]
            direct_kpoint = bs.kpoints[direct_kindex].frac_coords
            direct_kpoint = kpt_str.format(k=direct_kpoint)
            eq_kpoints = bs.get_equivalent_kpoints(direct_kindex)
            k_indices = ", ".join(map(str, eq_kpoints))

            # add 1 to band indices to be consistent with VASP band numbers.
            b_indices = ", ".join([str(i + 1) for i in spin_data["band_indices"]])

            click.echo(f"  {spin.name.capitalize()}:")
            click.echo(f"    k-point: {direct_kpoint}")
            click.echo(f"    k-point indices: {k_indices}")
            click.echo(f"    Band indices: {b_indices}")

    else:
        direct_bg = direct_data[Spin.up]["value"]
        click.echo(f"Direct band gap: {direct_bg:.3f} eV")

        direct_kindex = direct_data[Spin.up]["kpoint_index"]
        direct_kpoint = kpt_str.format(k=bs.kpoints[direct_kindex].frac_coords)
        k_indices = ", ".join(map(str, bs.get_equivalent_kpoints(direct_kindex)))
        b_indices = ", ".join(
            [str(i + 1) for i in direct_data[Spin.up]["band_indices"]]
        )

        click.echo(f"  k-point: {direct_kpoint}")
        click.echo(f"  k-point indices: {k_indices}")
        click.echo(f"  Band indices: {b_indices}")


def _log_band_edge_information(bs, edge_data):
    from pymatgen.electronic_structure.core import Spin

    if bs.is_spin_polarized:
        spins = edge_data["band_index"].keys()
        b_indices = [
            ", ".join([str(i + 1) for i in edge_data["band_index"][spin]])
            + f"({spin.name.capitalize()})"
            for spin in spins
        ]
        b_indices = ", ".join(b_indices)
    else:
        b_indices = ", ".join([str(i + 1) for i in edge_data["band_index"][Spin.up]])

    kpoint = edge_data["kpoint"]
    kpoint_str = kpt_str.format(k=kpoint.frac_coords)
    k_indices = ", ".join(map(str, edge_data["kpoint_index"]))

    if kpoint.label:
        k_loc = kpoint.label
    else:
        branch = bs.get_branch(edge_data["kpoint_index"][0])[0]
        k_loc = "between {}".format(branch["name"])

    click.echo("  Energy: {:.3f} eV".format(edge_data["energy"]))
    click.echo(f"  k-point: {kpoint_str}")
    click.echo(f"  k-point location: {k_loc}")
    click.echo(f"  k-point indices: {k_indices}")
    click.echo(f"  Band indices: {b_indices}")


def _log_effective_mass_data(data, is_spin_polarized, mass_type="m_e"):
    s = " ({})".format(data["spin"].name) if is_spin_polarized else ""

    # add 1 to band id to be consistent with VASP
    band_str = "band {}{}".format(data["band_id"] + 1, s)

    start_kpoint = data["start_kpoint"]
    end_kpoint = data["end_kpoint"]
    eff_mass = data["effective_mass"]

    kpoint_str = kpt_str.format(k=start_kpoint.frac_coords)
    if start_kpoint.label:
        kpoint_str += f" ({start_kpoint.label})"
    kpoint_str += " -> "
    kpoint_str += kpt_str.format(k=end_kpoint.frac_coords)
    if end_kpoint.label:
        kpoint_str += f" ({end_kpoint.label})"

    click.echo(f"  {mass_type}: {eff_mass:.3f} | {band_str} | {kpoint_str}")



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, Tuple

import numpy as np
from pymatgen.electronic_structure.core import Spin
from tabulate import tabulate

from amset.constants import (
    boltzmann_au,
    coulomb_to_au,
    ev_to_hartree,
    gpa_to_au,
    m_to_bohr,
    s_to_au,
)
from amset.core.data import AmsetData, check_nbands_equal
from amset.interpolation.deformation import DeformationPotentialInterpolator
from amset.scattering.common import calculate_inverse_screening_length_sq

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class AbstractElasticScattering(ABC):

    name: str
    required_properties: Tuple[str]

    def __init__(self, properties, doping, temperatures, nbands):
        self.properties = properties
        self.doping = doping
        self.temperatures = temperatures
        self.nbands = nbands
        self.spins = list(nbands.keys())

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
        )

    @abstractmethod
    def prefactor(self, spin: Spin, b_idx: int):
        pass

    @abstractmethod
    def factor(
        self,
        unit_q: np.array,
        norm_q_sq: np.ndarray,
        spin: Spin,
        band_idx: int,
        kpoint: np.ndarray,
        velocity: np.ndarray,
    ):
        pass

    def to_reference(self):
        return self.properties, self.doping, self.temperatures, self.nbands

    @classmethod
    def from_reference(cls, properties, doping, temperatures, nbands):
        return cls(properties, doping, temperatures, nbands)

    @classmethod
    def get_properties(cls, materials_properties):
        return {p: materials_properties[p] for p in cls.required_properties}

    @staticmethod
    def get_nbands(amset_data):
        return {s: len(amset_data.energies[s]) for s in amset_data.spins}


class AcousticDeformationPotentialScattering(AbstractElasticScattering):

    name = "ADP"
    required_properties = ("deformation_potential", "elastic_constant")

    def __init__(
        self,
        properties,
        doping,
        temperatures,
        nbands,
        deformation_potential,
        vb_idx,
        is_metal,
        fermi_levels,
    ):
        super().__init__(properties, doping, temperatures, nbands)
        self._prefactor = boltzmann_au * s_to_au
        self.elastic_constant = self.properties["elastic_constant"] * gpa_to_au
        self.deformation_potential = deformation_potential
        self.vb_idx = vb_idx
        self.is_metal = is_metal
        self.fermi_levels = fermi_levels

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        vb_idx = amset_data.vb_idx
        is_metal = amset_data.is_metal

        deformation_potential = materials_properties["deformation_potential"]
        if isinstance(deformation_potential, (str, Path)):
            deformation_potential = DeformationPotentialInterpolator.from_file(
                deformation_potential, scale=ev_to_hartree
            )
            equal = check_nbands_equal(deformation_potential, amset_data)
            if not equal:
                raise RuntimeError(
                    "Deformation potential file does not contain the correct number of"
                    " bands\nEnsure it was generated using the same energy_cutoff as "
                    "this AMSET run."
                )

        elif is_metal and isinstance(deformation_potential, tuple):
            logger.warning(
                "System is metallic but deformation potentials for both "
                "the valence and conduction bands have been set... using the "
                "valence band potential for all bands"
            )
            deformation_potential = deformation_potential[0] * ev_to_hartree

        elif is_metal:
            deformation_potential = deformation_potential * ev_to_hartree

        elif not is_metal and not isinstance(deformation_potential, tuple):
            logger.warning(
                "System is semiconducting but only one deformation "
                "potential has been set... using this potential for all bands."
            )
            deformation_potential = (
                deformation_potential * ev_to_hartree,
                deformation_potential * ev_to_hartree,
            )

        else:
            deformation_potential = (
                deformation_potential[0] * ev_to_hartree,
                deformation_potential[1] * ev_to_hartree,
            )
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
            deformation_potential,
            vb_idx,
            is_metal,
            amset_data.fermi_levels,
        )

    def prefactor(self, spin: Spin, b_idx: int):
        prefactor = (
            self._prefactor
            * self.temperatures[None, :]
            * np.ones((len(self.doping), len(self.temperatures)))
        )
        return prefactor

    def factor(
        self,
        unit_q: np.array,
        norm_q_sq: np.ndarray,
        spin: Spin,
        band_idx: int,
        kpoint: np.ndarray,
        velocity: np.ndarray,
    ):
        christoffel_tensors = get_christoffel_tensors(self.elastic_constant, unit_q)
        (
            (c_trans_a, c_trans_b, c_long),
            (v_trans_a, v_trans_b, v_long),
        ) = solve_christoffel_equation(christoffel_tensors)
        if isinstance(self.deformation_potential, DeformationPotentialInterpolator):
            deform = self.deformation_potential.interpolate(spin, [band_idx], [kpoint])
            deform = np.abs(deform[0])
            deform += np.outer(velocity, velocity)  # velocity correction
            strain_long, strain_trans_a, strain_trans_b = prepare_acoustic_strains(
                unit_q, v_long, v_trans_a, v_trans_b
            )
            factor = (
                np.tensordot(strain_long, deform) ** 2 / c_long
                + np.tensordot(strain_trans_a, deform) ** 2 / c_trans_a
                + np.tensordot(strain_trans_b, deform) ** 2 / c_trans_b
            )
        elif self.is_metal:
            factor = self.deformation_potential ** 2 / c_long
        else:
            def_idx = 1 if band_idx > self.vb_idx[spin] else 0
            factor = self.deformation_potential[def_idx] ** 2 / c_long

        return factor[None, None] * np.ones(self.fermi_levels.shape + norm_q_sq.shape)

    def to_reference(self):
        base_reference = super().to_reference()
        if isinstance(self.deformation_potential, DeformationPotentialInterpolator):
            deformation_reference = self.deformation_potential.to_reference()
            is_interpolator = True
        else:
            deformation_reference = self.deformation_potential
            is_interpolator = False
        return base_reference + (
            deformation_reference,
            self.vb_idx,
            self.is_metal,
            self.fermi_levels,
            is_interpolator,
        )

    @classmethod
    def from_reference(
        cls,
        properties,
        doping,
        temperatures,
        nbands,
        deformation_reference,
        vb_idx,
        is_metal,
        fermi_levels,
        is_interpolator,
    ):
        if is_interpolator:
            deformation_potential = DeformationPotentialInterpolator.from_reference(
                *deformation_reference
            )
        else:
            deformation_potential = deformation_reference
        return cls(
            properties,
            doping,
            temperatures,
            nbands,
            deformation_potential,
            vb_idx,
            is_metal,
            fermi_levels,
        )


def prepare_acoustic_strains(unit_q, v_long, v_trans_a, v_trans_b):
    # orient v_long and unit_q to face the same direction
    # the einsum is just pairwise dot product along the first axis
    sign = np.sign(np.einsum("ij,ij->i", unit_q, v_long))[:, None]
    v_long *= sign
    v_trans_a *= sign
    v_trans_b *= sign

    strain_long = get_unit_strain_tensors(unit_q, v_long)
    strain_trans_a = get_unit_strain_tensors(unit_q, v_trans_a)
    strain_trans_b = get_unit_strain_tensors(unit_q, v_trans_b)
    return strain_long, strain_trans_a, strain_trans_b


def get_christoffel_tensors(elastic_constant, unit_q):
    return np.einsum("ijkl,ni,nl->njk", elastic_constant, unit_q, unit_q)


def solve_christoffel_equation(christoffel_tensors):
    eigenvalues, eigenvectors = np.linalg.eigh(christoffel_tensors)
    return eigenvalues.T, eigenvectors.transpose(2, 0, 1)


def get_unit_strain_tensors(propagation_vectors, polarization_vectors):
    return propagation_vectors[:, :, None] * polarization_vectors[:, None, :]


class IonizedImpurityScattering(AbstractElasticScattering):
    name = "IMP"
    required_properties = ("defect_charge", "static_dielectric", "compensation_factor")

    def __init__(
        self,
        properties,
        doping,
        temperatures,
        nbands,
        q_concentration,
        inverse_screening_length_sq,
    ):
        super().__init__(properties, doping, temperatures, nbands)
        self._prefactor = q_concentration * s_to_au
        self.inverse_screening_length_sq = inverse_screening_length_sq

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        from amset.constants import bohr_to_cm

        avg_diel = np.linalg.eigvalsh(materials_properties["static_dielectric"]).mean()
        inverse_screening_length_sq = calculate_inverse_screening_length_sq(
            amset_data, avg_diel
        )
        defect_charge = materials_properties["defect_charge"]
        comp_factor = materials_properties["compensation_factor"]

        imp_info = []
        q_concentration = np.zeros(amset_data.fermi_levels.shape)
        for n, t in np.ndindex(inverse_screening_length_sq.shape):
            n_conc = np.abs(amset_data.electron_conc[n, t])
            p_conc = np.abs(amset_data.hole_conc[n, t])

            excess_impurities = abs((n_conc - p_conc) / defect_charge)
            impurity_concentration = excess_impurities * comp_factor
            q_concentration[n, t] = defect_charge ** 2 * impurity_concentration

            imp_info.append(
                (
                    amset_data.doping[n] * (1 / bohr_to_cm) ** 3,
                    amset_data.temperatures[t],
                    inverse_screening_length_sq[n, t],
                    impurity_concentration * (1 / bohr_to_cm) ** 3,
                )
            )

        logger.info("Inverse screening length (β) and impurity concentration (Nᵢᵢ):")
        table = tabulate(
            imp_info,
            headers=("conc [cm⁻³]", "temp [K]", "β² [a₀⁻²]", "Nᵢᵢ [cm⁻³]"),
            numalign="right",
            stralign="center",
            floatfmt=(".2e", ".1f", ".2e", ".2e"),
        )
        logger.info(table)
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
            q_concentration,
            inverse_screening_length_sq,
        )

    def prefactor(self, spin: Spin, b_idx: int):
        # need to return prefactor with shape (nspins, ndops, ntemps, nbands)
        return self._prefactor

    def factor(
        self,
        unit_q: np.array,
        norm_q_sq: np.ndarray,
        spin: Spin,
        band_idx: int,
        kpoint: np.ndarray,
        velocity: np.ndarray,
    ):
        static_tensor = self.properties["static_dielectric"] / (4 * np.pi)
        static_diel = np.einsum("ij,ij->i", unit_q, np.dot(static_tensor, unit_q.T).T)
        diel_factor = (1 / static_diel) ** 2
        return (
            diel_factor[None, None]
            / (norm_q_sq[None, None] + self.inverse_screening_length_sq[..., None]) ** 2
        )


class PiezoelectricScattering(AbstractElasticScattering):

    name = "PIE"
    required_properties = (
        "piezoelectric_constant",
        "elastic_constant",
        "high_frequency_dielectric",
        "free_carrier_screening",
    )

    def __init__(
        self,
        properties,
        doping,
        temperatures,
        nbands,
        piezoelectric_constant,
        inverse_screening_length_sq,
    ):
        super().__init__(properties, doping, temperatures, nbands)
        self.piezoelectric_constant = piezoelectric_constant
        self.inverse_screening_length_sq = inverse_screening_length_sq

        self._prefactor = self.temperatures[None, :] * boltzmann_au * s_to_au
        self._shape = np.ones((len(self.doping), len(self.temperatures)))
        self.elastic_constant = self.properties["elastic_constant"] * gpa_to_au

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        # convert dielectric to atomic units
        shape = (len(amset_data.doping), len(amset_data.temperatures))
        e = materials_properties["piezoelectric_constant"]
        e *= coulomb_to_au / m_to_bohr ** 2  # convert to atomic units
        dielectric = materials_properties["high_frequency_dielectric"] / (4 * np.pi)
        inv_dielectric = np.linalg.inv(dielectric)

        # use h piezoelectric coefficient (Stress-Voltage)
        piezoelectric_constant = np.einsum("mn,mkl->nkl", inv_dielectric, e)

        if materials_properties["free_carrier_screening"]:
            avg_diel = np.linalg.eigvalsh(
                materials_properties["high_frequency_dielectric"]
            ).mean()
            inverse_screening_length_sq = calculate_inverse_screening_length_sq(
                amset_data, avg_diel
            )
        else:
            # fill with small value for numerical convergence
            inverse_screening_length_sq = np.full(shape, 1e-12)
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
            piezoelectric_constant,
            inverse_screening_length_sq,
        )

    def prefactor(self, spin: Spin, b_idx: int):
        # need to return prefactor with shape (ndops, ntemps)
        return self._prefactor * self._shape

    def factor(
        self,
        unit_q: np.array,
        norm_q_sq: np.ndarray,
        spin: Spin,
        band_idx: int,
        kpoint: np.ndarray,
        velocity: np.ndarray,
    ):
        christoffel_tensors = get_christoffel_tensors(self.elastic_constant, unit_q)
        (
            (c_trans_a, c_trans_b, c_long),
            (v_trans_a, v_trans_b, v_long),
        ) = solve_christoffel_equation(christoffel_tensors)

        strain_long, strain_trans_a, strain_trans_b = prepare_acoustic_strains(
            unit_q, v_long, v_trans_a, v_trans_b
        )
        qh = np.einsum("ijk,nj->nik", self.piezoelectric_constant, unit_q)

        # einsum is double dot product along first axis
        factor = (
            np.einsum("nij,nij->n", strain_long, qh) ** 2 / c_long
            + np.einsum("nij,nij->n", strain_trans_a, qh) ** 2 / c_trans_a
            + np.einsum("nij,nij->n", strain_trans_b, qh) ** 2 / c_trans_b
        )

        return (
            factor[None, None]
            * np.ones(self._shape.shape + norm_q_sq.shape)
            / (norm_q_sq[None, None, :] + self.inverse_screening_length_sq[..., None])
        )



import logging

import numpy as np
from pymatgen.core.operations import SymmOp
from pymatgen.core.structure import Structure
from pymatgen.electronic_structure.bandstructure import BandStructure

from amset.constants import defaults
from amset.electronic_structure.kpoints import (
    get_kpoints_from_bandstructure,
    get_mesh_from_kpoint_diff,
    kpoints_to_first_bz,
    ktol,
    sort_kpoints,
)
from amset.log import log_list

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


logger = logging.getLogger(__name__)


def get_symmetry_equivalent_kpoints(
    structure,
    kpoints,
    symprec=defaults["symprec"],
    tol=ktol,
    return_inverse=False,
    time_reversal_symmetry=True,
):
    round_dp = int(np.log10(1 / tol))

    def shift_and_round(k):
        k = kpoints_to_first_bz(k)
        k = np.round(k, round_dp)
        return list(map(tuple, k))

    kpoints = np.asarray(kpoints)
    round_kpoints = shift_and_round(kpoints)

    rotation_matrices, _, _ = get_reciprocal_point_group_operations(
        structure, symprec=symprec, time_reversal=time_reversal_symmetry
    )

    equiv_points_mapping = {}
    rotation_matrix_mapping = {}
    mapping = []
    rot_mapping = []

    for i, point in enumerate(round_kpoints):

        if point in equiv_points_mapping:
            map_idx = equiv_points_mapping[point]
            mapping.append(map_idx)
            rot_mapping.append(rotation_matrix_mapping[map_idx][point])
        else:
            new_points = shift_and_round(np.dot(rotation_matrices, kpoints[i]))

            equiv_points_mapping.update(zip(new_points, [i] * len(new_points)))
            rotation_matrix_mapping[i] = dict(zip(new_points, rotation_matrices))

            mapping.append(i)
            rot_mapping.append(np.eye(3))

    ir_kpoints_idx, ir_to_full_idx, weights = np.unique(
        mapping, return_inverse=True, return_counts=True
    )

    ir_kpoints = kpoints[ir_kpoints_idx]

    if return_inverse:
        return (
            ir_kpoints,
            weights,
            ir_kpoints_idx,
            ir_to_full_idx,
            np.array(mapping),
            np.array(rot_mapping),
        )
    else:
        return ir_kpoints, weights


def similarity_transformation(rot, mat):
    """R x M x R^-1"""
    return np.dot(rot, np.dot(mat, np.linalg.inv(rot)))


def expand_kpoints(
    structure,
    kpoints,
    symprec=defaults["symprec"],
    return_mapping=False,
    time_reversal=True,
    verbose=True,
):
    if verbose:
        logger.info("Desymmetrizing k-point mesh")

    kpoints = np.array(kpoints).round(8)

    # due to limited input precision of the k-points, the mesh is returned as a float
    mesh, is_shifted = get_mesh_from_kpoint_diff(kpoints)
    status_info = ["Found initial mesh: {:.3f} x {:.3f} x {:.3f}".format(*mesh)]

    if is_shifted:
        shift = np.array([1, 1, 1])
    else:
        shift = np.array([0, 0, 0])

    # to avoid issues to limited input precision, recalculate the input k-points
    # so that the mesh is integer and the k-points are not truncated
    # to a small precision
    addresses = np.rint((kpoints + shift / (mesh * 2)) * mesh)
    mesh = np.rint(mesh)
    kpoints = addresses / mesh - shift / (mesh * 2)

    status_info.append("Integer mesh: {} x {} x {}".format(*map(int, mesh)))

    rotations, translations, is_tr = get_reciprocal_point_group_operations(
        structure, symprec=symprec, time_reversal=time_reversal
    )
    n_ops = len(rotations)
    if verbose:
        status_info.append(f"Using {n_ops} symmetry operations")
        log_list(status_info)

    # rotate all-kpoints
    all_rotated_kpoints = []
    for r in rotations:
        all_rotated_kpoints.append(np.dot(r, kpoints.T).T)
    all_rotated_kpoints = np.concatenate(all_rotated_kpoints)

    # map to first BZ
    all_rotated_kpoints -= np.rint(all_rotated_kpoints)
    all_rotated_kpoints = all_rotated_kpoints.round(8)

    # zone boundary consistent with VASP not with spglib
    all_rotated_kpoints[all_rotated_kpoints == -0.5] = 0.5

    # Find unique points
    unique_rotated_kpoints, unique_idxs = np.unique(
        all_rotated_kpoints, return_index=True, axis=0
    )

    # find integer addresses
    unique_addresses = (unique_rotated_kpoints + shift / (mesh * 2)) * mesh
    unique_addresses -= np.rint(unique_addresses)
    in_uniform_mesh = (np.abs(unique_addresses) < 1e-5).all(axis=1)

    n_mapped = int(np.sum(in_uniform_mesh))
    n_expected = int(np.product(mesh))
    if n_mapped != n_expected:
        raise ValueError(f"Expected {n_expected} points but found {n_mapped}")

    full_kpoints = unique_rotated_kpoints[in_uniform_mesh]
    full_idxs = unique_idxs[in_uniform_mesh]

    if not return_mapping:
        return full_kpoints

    op_mapping = np.floor(full_idxs / len(kpoints)).astype(int)
    kp_mapping = (full_idxs % len(kpoints)).astype(int)

    return full_kpoints, rotations, translations, is_tr, op_mapping, kp_mapping


def get_reciprocal_point_group_operations(
    structure: Structure,
    symprec: float = defaults["symprec"],
    time_reversal: bool = True,
):
    from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

    sga = SpacegroupAnalyzer(structure, symprec=symprec)
    if sga.get_symmetry_dataset() is None:
        # sometimes default angle tolerance doesn't work as expected
        sga = SpacegroupAnalyzer(structure, symprec=symprec, angle_tolerance=-1)

    rotations = sga.get_symmetry_dataset()["rotations"].transpose((0, 2, 1))
    translations = sga.get_symmetry_dataset()["translations"]
    is_tr = np.full(len(rotations), False, dtype=bool)

    if time_reversal:
        rotations = np.concatenate([rotations, -rotations])
        translations = np.concatenate([translations, -translations])
        is_tr = np.concatenate([is_tr, ~is_tr])

        rotations, unique_ops = np.unique(rotations, axis=0, return_index=True)
        translations = translations[unique_ops]
        is_tr = is_tr[unique_ops]

    # put identity first and time-reversal last
    sort_idx = np.argsort(np.abs(rotations - np.eye(3)).sum(axis=(1, 2)) + is_tr * 10)

    return rotations[sort_idx], translations[sort_idx], is_tr[sort_idx]


def expand_bandstructure(
    bandstructure, symprec=defaults["symprec"], time_reversal=True
):
    kpoints = get_kpoints_from_bandstructure(bandstructure)
    full_kpoints, _, _, _, _, kp_mapping = expand_kpoints(
        bandstructure.structure,
        kpoints,
        symprec=symprec,
        time_reversal=time_reversal,
        return_mapping=True,
    )
    return BandStructure(
        full_kpoints,
        {s: b[:, kp_mapping] for s, b in bandstructure.bands.items()},
        bandstructure.structure.lattice.reciprocal_lattice,
        bandstructure.efermi,
        structure=bandstructure.structure,
    )


def rotate_bandstructure(bandstructure: BandStructure, frac_symop: SymmOp):
    """Won't rotate projections..."""
    kpoints = get_kpoints_from_bandstructure(bandstructure)
    recip_rot = frac_symop.rotation_matrix
    rot_kpoints = np.dot(recip_rot, kpoints.T).T

    # map to first BZ, use VASP zone boundary convention
    rot_kpoints = kpoints_to_first_bz(rot_kpoints, negative_zone_boundary=False)

    # rotate structure
    structure = bandstructure.structure.copy()
    structure.apply_operation(frac_symop, fractional=True)

    return BandStructure(
        rot_kpoints,
        bandstructure.bands,
        structure.lattice.reciprocal_lattice,
        bandstructure.efermi,
        structure=structure,
    )


def get_symmops(structure, symprec=defaults["symprec"]):
    """Returns fractional ops as the cartesian ones from pmg have issues."""
    from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

    sga = SpacegroupAnalyzer(structure, symprec=symprec)

    # fractional rotation matrices from spglib need to be transposed so that the
    # operation is R.M
    return [
        SymmOp.from_rotation_and_translation(
            rotation_matrix=o.rotation_matrix.T, translation_vec=o.translation_vector
        )
        for o in sga.get_symmetry_operations(cartesian=False)
    ]


def reciprocal_lattice_match(
    band_structure: BandStructure, symprec=defaults["symprec"], tol=ktol
):
    """Test whether reciprocal lattice and k-point mesh are consistent.

    I.e., do all the reciprocal space group operations result preserve the k-point mesh.
    """
    round_dp = int(np.log10(1 / tol))

    kpoints = get_kpoints_from_bandstructure(band_structure)
    kpoints = expand_kpoints(
        band_structure.structure, kpoints, symprec=symprec, verbose=False
    )
    rotations, _, _ = get_reciprocal_point_group_operations(
        band_structure.structure, symprec=symprec
    )
    kpoints = kpoints_to_first_bz(kpoints.round(round_dp))
    kpoints = sort_kpoints(kpoints)

    for rotation in rotations:
        rot_kpoints = np.dot(rotation, kpoints.T).T
        rot_kpoints = kpoints_to_first_bz(rot_kpoints).round(round_dp)
        rot_kpoints = sort_kpoints(rot_kpoints)
        kpoints_match = np.max(np.linalg.norm(rot_kpoints - kpoints, axis=1)) < 0.01

        if not kpoints_match:
            return False

    return True


def rotation_matrix_to_su2(rotation_matrix, eps=1e-8):
    """Take a rotation matrix and turn it into a matrix that represents the
    same rotation in spin space [SU(2) matrix].

    Adapted from the `find_u` function in Quantum Espresso.

    Args:
        rotation_matrix: A 3x3 rotation matrix.
        eps: Numerical tolerance.

    Returns:
        The rotation matrix in SU(2) form.
    """
    det = np.linalg.det(rotation_matrix)
    rotation_matrix = np.array(rotation_matrix)

    # inversion has no effect in spin space, so improper rotations are multiplied by
    # inversion
    if np.abs(det + 1) < eps:
        saux = -rotation_matrix
    else:
        saux = rotation_matrix

    # Check for identity or inversion
    if np.allclose(saux, np.eye(3), atol=eps):
        return np.array([[1, 0], [0, 1]], dtype=np.complex)

    # Find the rotation axis and the rotation angle
    ax = get_rotation_axis(saux)
    angle = get_rotation_angle(saux)
    angle *= 0.5 * np.pi / 180
    cosa = np.cos(angle)
    sina = np.sin(angle)

    # Set the spin space rotation matrix elements
    u11 = complex(cosa, -ax[2] * sina)
    u12 = complex(-ax[1] * sina, -ax[0] * sina)
    u = np.array([[u11, u12], [u12.conjugate(), u11.conjugate()]], dtype=np.complex)

    # For each 3x3 rotation one can associate two 2x2 rotation matrices in spin
    # space. This function returns the U matrix with positive cosa term
    if cosa < -eps:
        u = -u
    return u


def get_rotation_axis(rotation_matrix, eps=1e-7):
    """
    Find the rotation axis from a rotation matrix.

    The orientation of the axis is with the tip in the hemisphere z >= 0. In the xy
    plane the axis is in the x > 0 region and the positive y axis is taken for z = 0
    and x = 0.

    Adapted from the `versor` subroutine in QE.

    Args:
        rotation_matrix: A 3x3 rotation matrix.
        eps: Numerical tolerance.

    Returns:
        A 3x1 vector of the rotation axis.
    """
    ts = get_symmetry_type(rotation_matrix)
    rotation_matrix = np.array(rotation_matrix)

    if ts not in (3, 4, 6):
        raise ValueError("Transformation matrix is not a proper or improper rotation.")

    a1 = np.zeros(3)
    if ts == 4:
        # Proper rotation of 180 degrees

        # First, the case where the axis is parallel to a coordinate axis
        ax = np.zeros(3)
        for i in range(3):
            if abs(rotation_matrix[i, i] - 1) < eps:
                ax[i] = 1
        norm = np.linalg.norm(ax)

        if abs(norm) > eps:
            return ax

        # Then the general case
        for i in range(3):
            a1[i] = np.sqrt(np.abs(rotation_matrix[i, i] + 1) / 2)

        for i in range(3):
            for j in range(i + 1, 3):
                if np.abs(a1[i] * a1[j]) > eps:
                    a1[i] = 0.5 * rotation_matrix[i, j] / a1[j]
    else:
        # It is not a 180 rotation: compute the rotation axis
        a1[0] = -rotation_matrix[1, 2] + rotation_matrix[2, 1]
        a1[1] = -rotation_matrix[2, 0] + rotation_matrix[0, 2]
        a1[2] = -rotation_matrix[0, 1] + rotation_matrix[1, 0]

    # The direction of the axis is arbitrarily chosen, with positive z. In the
    # xy plane with positive x, and along y with positive y.
    if a1[2] < -eps:
        a1 = -a1
    elif abs(a1[2]) < eps and a1[0] < -eps:
        a1 = -a1
    elif abs(a1[2]) < eps and abs(a1[0]) < eps and a1[1] < -eps:
        a1 = -a1

    norm = np.linalg.norm(a1)
    if norm < eps:
        raise ValueError("Problem with rotation matrix")
    return a1 / norm


def get_rotation_angle(rotation_matrix, eps=1e-7):
    """Get the rotation angle from a rotation matrix

    Args:
        rotation_matrix: A 3x3 rotation matrix.
        eps: Numerical tolerance.

    Returns:
        The rotation angle.
    """
    if get_symmetry_type(rotation_matrix) == 4:
        # proper 180 rotation
        return 180

    rotation_matrix = np.array(rotation_matrix)

    # compute the axis
    a1 = np.zeros(3)
    a1[0] = -rotation_matrix[1, 2] + rotation_matrix[2, 1]
    a1[1] = -rotation_matrix[2, 0] + rotation_matrix[0, 2]
    a1[2] = -rotation_matrix[0, 1] + rotation_matrix[1, 0]

    sint = 0.5 * np.linalg.norm(a1)
    if sint < eps or abs(sint) > 1 + eps:
        raise ValueError("Invalid rotation matrix")

    # small rounding errors that make |sint|>1.0 produce NaN in the next arcsin
    # function, so we remove them
    if abs(sint) > 1:
        sint = np.sign(sint)

    # The direction of the axis is arbitrarily chosen, with positive z. In the
    # xy plane with positive x, and along y with positive y.
    ax = a1.copy()
    if ax[2] < -eps:
        ax = -ax
    elif abs(ax[2]) < eps and ax[1] < -eps:
        ax = -ax
    elif abs(ax[2]) < eps and abs(ax[1]) < eps and ax[0] < -eps:
        ax = -ax

    if abs(a1[0]) > eps:
        sint = abs(sint) * np.sign(a1[0] / ax[0])
    elif abs(a1[1]) > eps:
        sint = abs(sint) * np.sign(a1[1] / ax[1])
    elif abs(a1[2]) > eps:
        sint = abs(sint) * np.sign(a1[2] / ax[2])

    # Compute the cos of the angle
    ax = a1 / (2 * sint)
    cost = 0
    if abs(ax[0] ** 2 - 1) > eps:
        cost = (rotation_matrix[0, 0] - ax[0] ** 2) / (1 - ax[0] ** 2)
    elif abs(ax[1] ** 2 - 1) > eps:
        cost = (rotation_matrix[1, 1] - ax[1] ** 2) / (1 - ax[1] ** 2)
    elif abs(ax[2] ** 2 - 1) > eps:
        cost = (rotation_matrix[2, 2] - ax[2] ** 2) / (1 - ax[2] ** 2)

    if abs(sint ** 2 + cost ** 2 - 1) > eps:
        raise ValueError("Problem calculating rotation angle.")

    angle_rot1 = np.arcsin(sint) * 180 / np.pi
    if angle_rot1 < 0:
        if cost < 0:
            angle_rot1 = -angle_rot1 + 180
        else:
            angle_rot1 = 360 + angle_rot1
    else:
        if cost < 0:
            angle_rot1 = -angle_rot1 + 180

    return angle_rot1


def get_symmetry_type(symmetry_matrix, eps=1e-7):
    """This function receives a 3x3 orthogonal matrix which is a symmetry
    operation of the point group of the crystal written in cartesian
    coordinates and gives as output a code according to the following:

    1. identity
    2. inversion
    3. proper rotation of an angle <> 180 degrees
    4. proper rotation of 180 degrees
    5. mirror symmetry
    6. improper rotation

    Args:
        symmetry_matrix: A 3x3 symmetry matrix.
        eps: Numerical tolerance.

    Returns:
        The symmetry type.
    """

    # check for identity
    if np.allclose(symmetry_matrix, np.eye(3), atol=eps):
        return 1

    # check for inversion
    if np.allclose(symmetry_matrix, -np.eye(3), atol=eps):
        return 2

    det = np.linalg.det(symmetry_matrix)

    # Determinant equal to 1: proper rotation
    if abs(det - 1) < eps:
        # check if an eigenvalue is equal to -1 (180 rotation)
        det1 = np.linalg.det(symmetry_matrix + np.eye(3))
        if abs(det1) < eps:
            # 180 proper rotation
            return 4
        else:
            # proper rotation <> 180
            return 3

    # Determinant equal to -1: mirror symmetry or improper rotation
    if abs(det + 1) < eps:
        # check if an eigenvalue is equal to 1 (mirror symmetry)
        det1 = np.linalg.det(symmetry_matrix - np.eye(3))

        if abs(det1) < eps:
            # mirror symmetry
            return 5
        else:
            # improper rotation
            return 6

    raise ValueError("Unrecognised symmetry")



import logging
import time
from typing import Dict, List, Optional, Union

import numpy as np
from pymatgen.core.structure import Structure
from pymatgen.electronic_structure.core import Spin
from pymatgen.util.coord import pbc_diff

from amset.constants import numeric_types
from amset.log import log_time_taken
from amset.util import (
    array_from_buffer,
    create_shared_array,
    create_shared_dict_array,
    dict_array_from_buffer,
    get_progress_bar,
    groupby,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

"""
     6-------7
    /|      /|
   / |     / |
  4-------5  |
  |  2----|--3
  | /     | /
  |/      |/
  0-------1

 i: vec        neighbours
 0: O          1, 2, 4
 1: a          0, 3, 5
 2: b          0, 3, 6
 3: a + b      1, 2, 7
 4: c          0, 5, 6
 5: c + a      1, 4, 7
 6: c + b      2, 4, 7
 7: c + a + b  3, 5, 6
"""

_main_diagonals = (
    (1, 1, 1),  # 0-7
    (-1, 1, 1),  # 1-6
    (1, -1, 1),  # 2-5
    (1, 1, -1),  # 3-4
)

_tetrahedron_vectors = np.array(
    [
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
        [1, 1, 0],
        [0, 0, 1],
        [1, 0, 1],
        [0, 1, 1],
        [1, 1, 1],
    ]
)

logger = logging.getLogger(__name__)


def get_main_diagonal(reciprocal_lattice: np.ndarray) -> int:
    # want a list of tetrahedra as (k1, k2, k3, k4); as per the Bloechl paper,
    # the order of the k-points is irrelevant and therefore they should be ordered
    # in increasing number
    # All tetrahedrons for a specific submesh will share one common diagonal.
    # To minimise interpolaiton distances, we should choose the shortest
    # diagonal
    diagonal_lengths = []
    for diagonal in _main_diagonals:
        length = np.linalg.norm(reciprocal_lattice @ diagonal)
        diagonal_lengths.append(length)

    return int(np.argmin(diagonal_lengths))


def get_relative_tetrahedron_vertices(reciprocal_lattice):
    shortest_index = get_main_diagonal(reciprocal_lattice)

    if shortest_index == 0:
        pairs = ((1, 3), (1, 5), (2, 3), (2, 6), (4, 5), (4, 6))
        main = (0, 7)

    elif shortest_index == 1:
        pairs = ((0, 2), (0, 4), (2, 3), (3, 7), (4, 5), (5, 7))
        main = (1, 6)

    elif shortest_index == 2:
        pairs = ((0, 1), (0, 4), (1, 3), (3, 7), (4, 6), (6, 7))
        main = (2, 5)

    elif shortest_index == 3:
        pairs = ((0, 1), (0, 2), (1, 5), (2, 6), (5, 7), (6, 7))
        main = (3, 4)

    else:
        assert False

    tetras = np.sort([main + x for x in pairs])
    return _tetrahedron_vectors[tetras]


def get_tetrahedra(
    reciprocal_lattice: np.ndarray,
    grid_address: np.ndarray,
    mesh: np.ndarray,
    grid_address_mapping,
):
    tetrahedron_vertices = get_relative_tetrahedron_vertices(reciprocal_lattice)

    grid_order = [1, mesh[0], mesh[0] * mesh[1]]

    all_grid_points = np.repeat(grid_address, [24] * len(grid_address), axis=0)
    all_vertices = np.tile(tetrahedron_vertices, (len(grid_address), 1, 1))
    points = all_grid_points.reshape(all_vertices.shape) + all_vertices

    # fancy magic from phonopy to get neighboring indices given relative coordinates
    tetrahedra = np.dot(points % mesh, grid_order)

    ir_tetrahedra_vertices = grid_address_mapping[tetrahedra]
    _, ir_tetrahedra_idx, ir_tetrahedra_to_full_idx, ir_weights = np.unique(
        np.sort(ir_tetrahedra_vertices),
        axis=0,
        return_index=True,
        return_inverse=True,
        return_counts=True,
    )

    return tetrahedra, ir_tetrahedra_idx, ir_tetrahedra_to_full_idx, ir_weights


class TetrahedralBandStructure:
    def __init__(
        self,
        energies: Dict[Spin, np.ndarray],
        kpoints: np.ndarray,
        ir_kpoints_idx: np.ndarray,
        ir_kpoint_mapping: np.ndarray,
        ir_kpoint_weights: np.ndarray,
        tetrahedra: Dict[Spin, np.ndarray],
        ir_tetrahedra: Dict[Spin, np.ndarray],
        ir_tetrahedra_energies: Dict[Spin, np.ndarray],
        ir_tetrahedra_idx: np.ndarray,
        ir_tetrahedra_to_full_idx: np.ndarray,
        ir_tetrahedra_weights: np.ndarray,
        e21: Dict[Spin, np.ndarray],
        e31: Dict[Spin, np.ndarray],
        e41: Dict[Spin, np.ndarray],
        e32: Dict[Spin, np.ndarray],
        e42: Dict[Spin, np.ndarray],
        e43: Dict[Spin, np.ndarray],
        max_tetrahedra_energies: Dict[Spin, np.ndarray],
        min_tetrahedra_energies: Dict[Spin, np.ndarray],
        cross_section_weights: Dict[Spin, np.ndarray],
        tetrahedron_volume: float,
        weights_cache: Optional[Dict[Spin, np.ndarray]] = None,
        weights_mask_cache: Optional[Dict[Spin, np.ndarray]] = None,
        energies_cache: Optional[Dict[Spin, np.ndarray]] = None,
    ):
        self.energies = energies
        self.kpoints = kpoints
        self.ir_kpoints_idx = ir_kpoints_idx
        self.ir_kpoint_mapping = ir_kpoint_mapping
        self.ir_kpoint_weights = ir_kpoint_weights
        self.tetrahedra = tetrahedra
        self.ir_tetrahedra = ir_tetrahedra
        self.ir_tetrahedra_energies = ir_tetrahedra_energies
        self.ir_tetrahedra_idx = ir_tetrahedra_idx
        self.ir_tetrahedra_to_full_idx = ir_tetrahedra_to_full_idx
        self.ir_tetrahedra_weights = ir_tetrahedra_weights
        self.e21 = e21
        self.e31 = e31
        self.e41 = e41
        self.e32 = e32
        self.e42 = e42
        self.e43 = e43
        self.max_tetrahedra_energies = max_tetrahedra_energies
        self.min_tetrahedra_energies = min_tetrahedra_energies
        self.cross_section_weights = cross_section_weights
        self._tetrahedron_volume = tetrahedron_volume
        self._weights_cache = {} if weights_cache is None else weights_cache
        self._weights_mask_cache = (
            {} if weights_mask_cache is None else weights_mask_cache
        )
        self._energies_cache = {} if energies_cache is None else energies_cache

        self.grouped_ir_to_full = groupby(
            np.arange(len(ir_tetrahedra_to_full_idx)), ir_tetrahedra_to_full_idx
        )
        self._ir_weights_shape = {
            s: (len(energies[s]), len(ir_kpoints_idx)) for s in energies
        }

    def to_reference(self):
        energies_buffer, self.energies = create_shared_dict_array(
            self.energies, return_shared_data=True
        )
        kpoints_buffer, self.kpoints = create_shared_array(
            self.kpoints, return_shared_data=True
        )
        ir_kpoints_idx_buffer, self.ir_kpoints_idx = create_shared_array(
            self.ir_kpoints_idx, return_shared_data=True
        )
        ir_kpoint_mapping_buffer, self.ir_kpoint_mapping = create_shared_array(
            self.ir_kpoint_mapping, return_shared_data=True
        )
        ir_kpoint_weights_buffer, self.ir_kpoint_weights = create_shared_array(
            self.ir_kpoint_weights, return_shared_data=True
        )
        tetrahedra_buffer, self.tetrahedra = create_shared_dict_array(
            self.tetrahedra, return_shared_data=True
        )
        ir_tetrahedra_buffer, self.ir_tetrahedra = create_shared_dict_array(
            self.ir_tetrahedra, return_shared_data=True
        )
        (
            ir_tetrahedra_energies_buffer,
            self.ir_tetrahedra_energies,
        ) = create_shared_dict_array(
            self.ir_tetrahedra_energies, return_shared_data=True
        )
        ir_tetrahedra_idx_buffer, self.ir_tetrahedra_idx = create_shared_array(
            self.ir_tetrahedra_idx, return_shared_data=True
        )
        (
            ir_tetrahedra_to_full_idx_buffer,
            self.ir_tetrahedra_to_full_idx,
        ) = create_shared_array(self.ir_tetrahedra_to_full_idx, return_shared_data=True)
        ir_tetrahedra_weights_buffer, self.ir_tetrahedra_weights = create_shared_array(
            self.ir_tetrahedra_weights, return_shared_data=True
        )
        e21_buffer, self.e21 = create_shared_dict_array(
            self.e21, return_shared_data=True
        )
        e31_buffer, self.e31 = create_shared_dict_array(
            self.e31, return_shared_data=True
        )
        e41_buffer, self.e41 = create_shared_dict_array(
            self.e41, return_shared_data=True
        )
        e32_buffer, self.e32 = create_shared_dict_array(
            self.e32, return_shared_data=True
        )
        e42_buffer, self.e42 = create_shared_dict_array(
            self.e42, return_shared_data=True
        )
        e43_buffer, self.e43 = create_shared_dict_array(
            self.e43, return_shared_data=True
        )
        (
            max_tetrahedra_energies_buffer,
            self.max_tetrahedra_energies,
        ) = create_shared_dict_array(
            self.max_tetrahedra_energies, return_shared_data=True
        )
        (
            min_tetrahedra_energies_buffer,
            self.min_tetrahedra_energies,
        ) = create_shared_dict_array(
            self.min_tetrahedra_energies, return_shared_data=True
        )
        (
            cross_section_weights_buffer,
            self.cross_section_weights,
        ) = create_shared_dict_array(
            self.cross_section_weights, return_shared_data=True
        )
        weights_cache_buffer, self._weights_cache = create_shared_dict_array(
            self._weights_cache, return_shared_data=True
        )
        weights_mask_cache_buffer, self._weights_mask_cache = create_shared_dict_array(
            self._weights_mask_cache, return_shared_data=True
        )
        energies_cache_buffer, self._energies_cache = create_shared_dict_array(
            self._energies_cache, return_shared_data=True
        )

        return (
            energies_buffer,
            kpoints_buffer,
            ir_kpoints_idx_buffer,
            ir_kpoint_mapping_buffer,
            ir_kpoint_weights_buffer,
            tetrahedra_buffer,
            ir_tetrahedra_buffer,
            ir_tetrahedra_energies_buffer,
            ir_tetrahedra_idx_buffer,
            ir_tetrahedra_to_full_idx_buffer,
            ir_tetrahedra_weights_buffer,
            e21_buffer,
            e31_buffer,
            e41_buffer,
            e32_buffer,
            e42_buffer,
            e43_buffer,
            max_tetrahedra_energies_buffer,
            min_tetrahedra_energies_buffer,
            cross_section_weights_buffer,
            self._tetrahedron_volume,
            weights_cache_buffer,
            weights_mask_cache_buffer,
            energies_cache_buffer,
        )

    @classmethod
    def from_reference(
        cls,
        energies_buffer,
        kpoints_buffer,
        ir_kpoints_idx_buffer,
        ir_kpoint_mapping_buffer,
        ir_kpoint_weights_buffer,
        tetrahedra_buffer,
        ir_tetrahedra_buffer,
        ir_tetrahedra_energies_buffer,
        ir_tetrahedra_idx_buffer,
        ir_tetrahedra_to_full_idx_buffer,
        ir_tetrahedra_weights_buffer,
        e21_buffer,
        e31_buffer,
        e41_buffer,
        e32_buffer,
        e42_buffer,
        e43_buffer,
        max_tetrahedra_energies_buffer,
        min_tetrahedra_energies_buffer,
        cross_section_weights_buffer,
        tetrahedron_volume,
        weights_cache_buffer,
        weights_mask_cache_buffer,
        energies_cache_buffer,
    ):
        return cls(
            dict_array_from_buffer(energies_buffer),
            array_from_buffer(kpoints_buffer),
            array_from_buffer(ir_kpoints_idx_buffer),
            array_from_buffer(ir_kpoint_mapping_buffer),
            array_from_buffer(ir_kpoint_weights_buffer),
            dict_array_from_buffer(tetrahedra_buffer),
            dict_array_from_buffer(ir_tetrahedra_buffer),
            dict_array_from_buffer(ir_tetrahedra_energies_buffer),
            array_from_buffer(ir_tetrahedra_idx_buffer),
            array_from_buffer(ir_tetrahedra_to_full_idx_buffer),
            array_from_buffer(ir_tetrahedra_weights_buffer),
            dict_array_from_buffer(e21_buffer),
            dict_array_from_buffer(e31_buffer),
            dict_array_from_buffer(e41_buffer),
            dict_array_from_buffer(e32_buffer),
            dict_array_from_buffer(e42_buffer),
            dict_array_from_buffer(e43_buffer),
            dict_array_from_buffer(max_tetrahedra_energies_buffer),
            dict_array_from_buffer(min_tetrahedra_energies_buffer),
            dict_array_from_buffer(cross_section_weights_buffer),
            tetrahedron_volume,
            dict_array_from_buffer(weights_cache_buffer),
            dict_array_from_buffer(weights_mask_cache_buffer),
            dict_array_from_buffer(energies_cache_buffer),
        )

    @classmethod
    def from_data(
        cls,
        energies: Dict[Spin, np.ndarray],
        kpoints: np.ndarray,
        tetrahedra: np.ndarray,
        structure: Structure,
        ir_kpoints_idx: np.ndarray,
        ir_kpoint_mapping: np.ndarray,
        ir_tetrahedra_idx: Optional[np.ndarray] = None,
        ir_tetrahedra_to_full_idx: Optional[np.ndarray] = None,
        ir_tetrahedra_weights: Optional[np.ndarray] = None,
    ):
        logger.info("Initializing tetrahedron band structure")
        t0 = time.perf_counter()

        tparams = (ir_tetrahedra_idx, ir_tetrahedra_to_full_idx, ir_tetrahedra_weights)
        if len({x is None for x in tparams}) != 1:
            raise ValueError(
                "Either all or none of ir_tetrahedra_idx, ir_tetrahedra_to_full_idx and"
                " ir_tetrahedra_weights should be set."
            )

        if ir_tetrahedra_idx is None:
            ir_tetrahedra_idx = np.arange(len(kpoints))
            ir_tetrahedra_to_full_idx = np.ones_like(ir_tetrahedra_idx)
            ir_tetrahedra_weights = np.ones_like(ir_tetrahedra_idx)

        ir_tetrahedra_to_full_idx = ir_tetrahedra_to_full_idx
        ir_kpoints_idx = ir_kpoints_idx
        ir_kpoint_mapping = ir_kpoint_mapping

        _, ir_kpoint_weights = np.unique(ir_kpoint_mapping, return_counts=True)

        # need to keep track of full tetrahedra to recover full k-point indices
        # when calculating scattering rates (i.e., k-k' is symmetry inequivalent).
        full_tetrahedra, _ = process_tetrahedra(tetrahedra, energies)

        # store irreducible tetrahedra and use energies to calculate diffs and min/maxes
        ir_tetrahedra, ir_tetrahedra_energies = process_tetrahedra(
            tetrahedra[ir_tetrahedra_idx], energies
        )

        # the remaining properties are given for each irreducible tetrahedra
        (e21, e31, e41, e32, e42, e43) = get_tetrahedra_energy_diffs(
            ir_tetrahedra_energies
        )

        (
            max_tetrahedra_energies,
            min_tetrahedra_energies,
        ) = get_max_min_tetrahedra_energies(ir_tetrahedra_energies)

        cross_section_weights = get_tetrahedra_cross_section_weights(
            structure.lattice.reciprocal_lattice.matrix,
            kpoints,
            ir_tetrahedra,
            e21,
            e31,
            e41,
        )

        tetrahedron_volume = 1 / len(tetrahedra)

        log_time_taken(t0)
        return cls(
            energies,
            kpoints,
            ir_kpoints_idx,
            ir_kpoint_mapping,
            ir_kpoint_weights,
            full_tetrahedra,
            ir_tetrahedra,
            ir_tetrahedra_energies,
            ir_tetrahedra_idx,
            ir_tetrahedra_to_full_idx,
            ir_tetrahedra_weights,
            e21,
            e31,
            e41,
            e32,
            e42,
            e43,
            max_tetrahedra_energies,
            min_tetrahedra_energies,
            cross_section_weights,
            tetrahedron_volume,
        )

    def get_connected_kpoints(self, kpoint_idx: Union[int, List[int], np.ndarray]):
        """Given one or more k-point indices, get a list of all k-points that are in
        the same tetrahedra

        Args:
            kpoint_idx: One or mode k-point indices.

        Returns:
            A list of k-point indices that are in the same tetrahedra.
        """
        if isinstance(kpoint_idx, numeric_types):
            kpoint_idx = [kpoint_idx]

        tetrahedra = self.tetrahedra[Spin.up][0]
        return np.unique(tetrahedra[np.isin(tetrahedra, kpoint_idx).any(axis=1)])

    def get_intersecting_tetrahedra(self, spin, energy, band_idx=None):

        max_energies = self.max_tetrahedra_energies[spin]
        min_energies = self.min_tetrahedra_energies[spin]

        if band_idx is not None:
            mask = np.full_like(max_energies, False, dtype=bool)
            mask[band_idx] = True

            return (min_energies < energy) & (max_energies > energy) & mask

        else:
            return (min_energies < energy) & (max_energies > energy)

    def get_tetrahedra_density_of_states(
        self,
        spin,
        energy,
        return_contributions=False,
        symmetry_reduce=True,
        band_idx=None,
    ):
        tetrahedra_mask = self.get_intersecting_tetrahedra(
            spin, energy, band_idx=band_idx
        )

        if not np.any(tetrahedra_mask):
            if return_contributions:
                return [], [], [], []
            else:
                return []

        energies = self.ir_tetrahedra_energies[spin][tetrahedra_mask]
        e21 = self.e21[spin][tetrahedra_mask]
        e31 = self.e31[spin][tetrahedra_mask]
        e41 = self.e41[spin][tetrahedra_mask]
        e32 = self.e32[spin][tetrahedra_mask]
        e42 = self.e42[spin][tetrahedra_mask]
        e43 = self.e43[spin][tetrahedra_mask]
        cs_weights = self.cross_section_weights[spin][tetrahedra_mask]

        cond_a_mask = (energies[:, 0] < energy) & (energy < energies[:, 1])
        cond_b_mask = (energies[:, 1] <= energy) & (energy < energies[:, 2])
        cond_c_mask = (energies[:, 2] <= energy) & (energy < energies[:, 3])

        ee1 = energy - energies[:, 0]
        ee2 = energy - energies[:, 1]
        e4e = energies[:, 3] - energy

        tetrahedra_dos = np.zeros(len(energies))

        tetrahedra_dos[cond_a_mask] = _get_density_of_states_a(
            ee1[cond_a_mask], e21[cond_a_mask], e31[cond_a_mask], e41[cond_a_mask]
        )

        tetrahedra_dos[cond_b_mask] = _get_density_of_states_b(
            ee2[cond_b_mask],
            e21[cond_b_mask],
            e31[cond_b_mask],
            e41[cond_b_mask],
            e32[cond_b_mask],
            e42[cond_b_mask],
        )

        tetrahedra_dos[cond_c_mask] = _get_density_of_states_c(
            e4e[cond_c_mask], e41[cond_c_mask], e42[cond_c_mask], e43[cond_c_mask]
        )

        tetrahedra_dos *= self._tetrahedron_volume

        band_idx, tetrahedra_idx = np.where(tetrahedra_mask)
        tetrahedra_weights = self.ir_tetrahedra_weights[tetrahedra_idx]

        if symmetry_reduce:
            tetrahedra_dos *= tetrahedra_weights

        else:
            # transform the mask to the full BZ
            band_idx = np.repeat(band_idx, tetrahedra_weights)
            tetrahedra_grouped = self.grouped_ir_to_full[tetrahedra_idx]
            tetrahedra_idx = np.concatenate(tetrahedra_grouped)
            tetrahedra_mask = (band_idx, tetrahedra_idx)

            # tetrahedra_mask = tetrahedra_mask[:, self.ir_tetrahedra_to_full_idx]
            # get a mask to expand the remaining properties with shape (ntetra,)
            # to the full BZ
            expand_tetrahedra = np.repeat(np.arange(len(ee1)), tetrahedra_weights)
            tetrahedra_dos = tetrahedra_dos[expand_tetrahedra]
            cond_a_mask = cond_a_mask[expand_tetrahedra]
            cond_b_mask = cond_b_mask[expand_tetrahedra]
            cond_c_mask = cond_c_mask[expand_tetrahedra]
            ee1 = ee1[expand_tetrahedra]
            ee2 = ee2[expand_tetrahedra]
            e4e = e4e[expand_tetrahedra]
            e21 = e21[expand_tetrahedra]
            e31 = e31[expand_tetrahedra]
            e41 = e41[expand_tetrahedra]
            e32 = e32[expand_tetrahedra]
            e42 = e42[expand_tetrahedra]
            e43 = e43[expand_tetrahedra]
            cs_weights = cs_weights[expand_tetrahedra]

        if return_contributions:
            frac_21 = ee1 / e21
            frac_31 = ee1 / e31
            frac_41 = ee1 / e41
            frac_32 = ee2 / e32
            frac_42 = ee2 / e42
            frac_c_41 = e4e / e41
            frac_c_42 = e4e / e42
            frac_c_43 = e4e / e43

            frac_21[(frac_21 == np.nan) | (frac_21 == np.inf)] = 1
            frac_31[(frac_31 == np.nan) | (frac_31 == np.inf)] = 1
            frac_41[(frac_41 == np.nan) | (frac_41 == np.inf)] = 1
            frac_32[(frac_32 == np.nan) | (frac_32 == np.inf)] = 1
            frac_42[(frac_42 == np.nan) | (frac_42 == np.inf)] = 1
            frac_c_41[(frac_c_41 == np.nan) | (frac_c_41 == np.inf)] = 1
            frac_c_42[(frac_c_42 == np.nan) | (frac_c_42 == np.inf)] = 1
            frac_c_43[(frac_c_43 == np.nan) | (frac_c_43 == np.inf)] = 1

            contributions = (
                cond_a_mask,
                cond_b_mask,
                cond_c_mask,
                frac_21,
                frac_31,
                frac_41,
                frac_32,
                frac_42,
                frac_c_41,
                frac_c_42,
                frac_c_43,
            )

            return tetrahedra_dos, tetrahedra_mask, cs_weights, contributions

        else:
            return tetrahedra_dos

    def get_density_of_states(
        self,
        energies=None,
        integrand=None,
        sum_spins=False,
        band_idx=None,
        use_cached_weights=False,
        progress_bar=False,
    ):
        if energies is None:
            from amset.constants import defaults, ev_to_hartree

            min_e = np.min([np.min(e) for e in self.energies.values()])
            max_e = np.max([np.max(e) for e in self.energies.values()])
            energies = np.arange(min_e, max_e, defaults["dos_estep"] * ev_to_hartree)

        dos = {}
        for spin in self.energies.keys():
            if isinstance(integrand, dict):
                # integrand given for each spin channel
                spin_integrand = integrand[spin]
            else:
                spin_integrand = integrand

            if isinstance(band_idx, dict):
                # band indices given for each spin channel
                spin_band_idx = band_idx[spin]
            else:
                spin_band_idx = band_idx

            if spin_integrand is not None:
                if spin_integrand.shape[:2] != self.energies[spin].shape:
                    raise ValueError(
                        "Unexpected integrand shape, should be (nbands, nkpoints, ...)"
                    )

                nbands = len(spin_integrand)
                integrand_shape = spin_integrand.shape[2:]
                n_ir_kpoints = len(self.ir_kpoints_idx)
                new_integrand = np.zeros((nbands, n_ir_kpoints) + integrand_shape)

                flat_k = np.tile(self.ir_kpoint_mapping, nbands)
                flat_b = np.repeat(np.arange(nbands), len(self.ir_kpoint_mapping))
                flat_integrand = spin_integrand.reshape((-1,) + integrand_shape)
                # flat_integrand = spin_integrand.reshape(-1, 3, 3)

                # sum integrand at all symmetry equivalent points, new_integrand
                # has shape (nbands, n_ir_kpoints)
                np.add.at(new_integrand, (flat_b, flat_k), flat_integrand)
                spin_integrand = new_integrand

            emesh, dos[spin] = self.get_spin_density_of_states(
                spin,
                energies,
                integrand=spin_integrand,
                band_idx=spin_band_idx,
                use_cached_weights=use_cached_weights,
                progress_bar=progress_bar,
            )

        if sum_spins:
            if Spin.down in dos:
                dos = dos[Spin.up] + dos[Spin.down]
            else:
                dos = dos[Spin.up]

        return energies, dos

    def get_spin_density_of_states(
        self,
        spin,
        energies,
        integrand=None,
        band_idx=None,
        use_cached_weights=False,
        progress_bar=False,
    ):
        # integrand should have the shape (nbands, n_ir_kpts, ...)
        # the integrand should have been summed at all equivalent k-points
        # TODO: add support for variable shaped integrands
        if integrand is None:
            dos = np.zeros_like(energies)
        else:
            integrand_shape = integrand.shape[2:]
            dos = np.zeros((len(energies),) + integrand_shape)

        if use_cached_weights:
            if self._weights_cache is None:
                raise ValueError("No integrand have been cached")

            all_weights = self._weights_cache[spin]
            all_weights_mask = self._weights_mask_cache[spin]
            energies = self._energies_cache[spin]
        else:
            all_weights = []
            all_weights_mask = []

        nbands = len(self.energies[spin])
        kpoint_multiplicity = np.tile(self.ir_kpoint_weights, (nbands, 1))

        if band_idx is not None and integrand is not None:
            integrand = integrand[band_idx]

        if band_idx is not None and integrand is None:
            kpoint_multiplicity = kpoint_multiplicity[band_idx]

        energies_iter = list(enumerate(energies))
        if progress_bar:
            energies_iter = get_progress_bar(iterable=energies_iter, desc="DOS")

        for i, energy in energies_iter:
            if use_cached_weights:
                weights = all_weights[i]
                weights_mask = all_weights_mask[i]
            else:
                weights = self.get_energy_dependent_integration_weights(spin, energy)
                weights_mask = weights != 0
                all_weights.append(weights)
                all_weights_mask.append(weights_mask)

            if band_idx is not None:
                weights = weights[band_idx]
                weights_mask = weights_mask[band_idx]

            if integrand is None:
                dos[i] = np.sum(
                    weights[weights_mask] * kpoint_multiplicity[weights_mask]
                )

            else:
                # expand weights to match the dimensions of the integrand
                expand_axis = [1 + i for i in range(len(integrand.shape[2:]))]
                expand_weights = np.expand_dims(weights[weights_mask], axis=expand_axis)

                # don't need to include the k-point multiplicity as this is included by
                # pre-summing the integrand at symmetry equivalent points
                dos[i] = np.sum(expand_weights * integrand[weights_mask], axis=0)

        if not use_cached_weights:
            self._weights_cache[spin] = np.array(all_weights)
            self._weights_mask_cache[spin] = np.array(all_weights_mask)
            self._energies_cache[spin] = energies

        return energies, np.asarray(dos)

    def get_energy_dependent_integration_weights(self, spin, energy):
        integration_weights = np.zeros(self._ir_weights_shape[spin])
        tetrahedra_mask = self.get_intersecting_tetrahedra(spin, energy)

        if not np.any(tetrahedra_mask):
            return integration_weights

        energies = self.ir_tetrahedra_energies[spin][tetrahedra_mask]
        e21 = self.e21[spin][tetrahedra_mask]
        e31 = self.e31[spin][tetrahedra_mask]
        e41 = self.e41[spin][tetrahedra_mask]
        e32 = self.e32[spin][tetrahedra_mask]
        e42 = self.e42[spin][tetrahedra_mask]
        e43 = self.e43[spin][tetrahedra_mask]

        cond_a_mask = (energies[:, 0] < energy) & (energy < energies[:, 1])
        cond_b_mask = (energies[:, 1] <= energy) & (energy < energies[:, 2])
        cond_c_mask = (energies[:, 2] <= energy) & (energy < energies[:, 3])

        ee1 = energy - energies[:, 0]
        ee2 = energy - energies[:, 1]
        ee3 = energy - energies[:, 2]
        e2e = energies[:, 1] - energy
        e3e = energies[:, 2] - energy
        e4e = energies[:, 3] - energy

        kpoints_idx = self.ir_tetrahedra[spin][tetrahedra_mask]
        ir_kpoints_idx = self.ir_kpoint_mapping[kpoints_idx]

        # calculate the integrand for each vertices
        vert_weights = np.zeros_like(energies)
        vert_weights[cond_a_mask] = _get_energy_dependent_weight_a(
            ee1[cond_a_mask],
            e2e[cond_a_mask],
            e3e[cond_a_mask],
            e4e[cond_a_mask],
            e21[cond_a_mask],
            e31[cond_a_mask],
            e41[cond_a_mask],
        )

        vert_weights[cond_b_mask] = _get_energy_dependent_weight_b(
            ee1[cond_b_mask],
            ee2[cond_b_mask],
            e3e[cond_b_mask],
            e4e[cond_b_mask],
            e31[cond_b_mask],
            e41[cond_b_mask],
            e32[cond_b_mask],
            e42[cond_b_mask],
        )

        vert_weights[cond_c_mask] = _get_energy_dependent_weight_c(
            ee1[cond_c_mask],
            ee2[cond_c_mask],
            ee3[cond_c_mask],
            e4e[cond_c_mask],
            e41[cond_c_mask],
            e42[cond_c_mask],
            e43[cond_c_mask],
        )

        # finally, get the integrand for each ir_kpoint by summing over all
        # tetrahedra and multiplying by the tetrahedra multiplicity and
        # tetrahedra weight; Finally, divide by the k-point multiplicity
        # to get the final weight
        band_idx, tetrahedra_idx = np.where(tetrahedra_mask)

        # include tetrahedra multiplicity
        vert_weights *= self.ir_tetrahedra_weights[tetrahedra_idx][:, None]

        flat_ir_kpoints = np.ravel(ir_kpoints_idx)
        flat_ir_weights = np.ravel(vert_weights)
        flat_bands = np.repeat(band_idx, 4)

        # sum integrand, note this sums in place and is insanely fast
        np.add.at(integration_weights, (flat_bands, flat_ir_kpoints), flat_ir_weights)
        integration_weights *= (
            self._tetrahedron_volume / self.ir_kpoint_weights[None, :]
        )

        return integration_weights

    def get_masks(self, spin, tetrahedra_mask):
        # mask needs to be generated with symmetry_reduce=False
        band_idxs = tetrahedra_mask[0]

        # property_mask can be used to get the values of a band and k-dependent
        # property at all tetrahedra vertices for each tetrahedron selected by the
        # mask for example: energies[property_mask]
        property_mask = (
            np.repeat(band_idxs[:, None], 4, axis=1),
            self.tetrahedra[spin][tetrahedra_mask],
        )

        # band_kpoint_mask can be used to get the inequivalent band, k-point
        # combinations that are required to get the properties for all tetrahedra
        # vertices. Can be used to avoid calculation of band and k-point dependent
        # properties at all bands k-points and k-points (instead you only calculate
        # the properties at the bands and k-points involved in the tetrahedra)
        # and therefore reduce computational expense
        band_kpoint_mask = np.full(self.energies[spin].shape, False)
        band_kpoint_mask[property_mask] = True

        # kpoint_mask is the k-point contribution to band_kpoint_mask which allows
        # mixing of band and k-point dependent properties and just k-dependent
        # properties
        band_mask, kpoint_mask = np.where(band_kpoint_mask)

        return property_mask, band_kpoint_mask, band_mask, kpoint_mask


def _get_density_of_states_a(ee1, e21, e31, e41):
    return 3 * ee1 ** 2 / (e21 * e31 * e41)


def _get_density_of_states_b(ee2, e21, e31, e41, e32, e42):
    return 3 * (e21 + 2 * ee2 - (e42 + e31) * ee2 ** 2 / (e32 * e42)) / (e31 * e41)


def _get_density_of_states_c(e4e, e41, e42, e43):
    return 3 * e4e ** 2 / (e41 * e42 * e43)


def _get_energy_dependent_weight_a(ee1, e2e, e3e, e4e, e21, e31, e41):
    c = ee1 ** 2 / (e21 * e31 * e41)
    i1 = c * (e2e / e21 + e3e / e31 + e4e / e41)
    i2 = c * (ee1 / e21)
    i3 = c * (ee1 / e31)
    i4 = c * (ee1 / e41)
    return np.stack([i1, i2, i3, i4], axis=1)


def _get_energy_dependent_weight_b(ee1, ee2, e3e, e4e, e31, e41, e32, e42):
    c = (ee1 * e4e) / (e31 * e41 * e42)
    x = e3e / e31
    y = e4e / e42
    z = ee2 / (e32 * e42)
    zx = z * x
    k = ee1 / e31
    n = ee2 / e42

    i1 = c * (x + e4e / e41) + z * x ** 2
    i2 = c * y + zx * (e3e / e32 + y)
    i3 = c * k + zx * (k + ee2 / e32)
    i4 = c * (ee1 / e41 + n) + zx * n
    return np.stack([i1, i2, i3, i4], axis=1)


def _get_energy_dependent_weight_c(ee1, ee2, ee3, e4e, e41, e42, e43):
    c = e4e ** 2 / (e41 * e42 * e43)
    i1 = c * e4e / e41
    i2 = c * e4e / e42
    i3 = c * e4e / e43
    i4 = c * (ee1 / e41 + ee2 / e42 + ee3 / e43)
    return np.stack([i1, i2, i3, i4], axis=1)


def get_cross_section_values(
    property_values,
    cond_a_mask,
    cond_b_mask,
    cond_c_mask,
    frac_21,
    frac_31,
    frac_41,
    frac_32,
    frac_42,
    frac_c_41,
    frac_c_42,
    frac_c_43,
    average=True,
):
    # property_values is given for each tetrahedra
    # property can be scalar or vector
    p21 = property_values[:, 1] - property_values[:, 0]
    p31 = property_values[:, 2] - property_values[:, 0]
    p41 = property_values[:, 3] - property_values[:, 0]
    p32 = property_values[:, 2] - property_values[:, 1]
    p42 = property_values[:, 3] - property_values[:, 1]
    p43 = property_values[:, 3] - property_values[:, 2]

    prop_shape = property_values.shape[2:]

    if average:
        # remove the 4 vertices from shape
        value_shape = property_values.shape[:1] + prop_shape
    else:
        # shape is (ntetrahedra, 4, prop_shape)
        # the 4 is to account for the intersections of the cross section and tetrahedra
        # the number of intersections can be either 3 or 4;
        # for cross sections with 3 interesections, the final intersection will be
        # an array of 0's with the shape of prop_shape
        value_shape = property_values.shape[:1] + (4,) + prop_shape

    values = np.zeros(value_shape)

    if len(prop_shape) != 0:
        # property is multi-dimensional
        new_shape = (len(values),) + tuple([1] * len(prop_shape))
        frac_21 = frac_21.reshape(new_shape)
        frac_31 = frac_31.reshape(new_shape)
        frac_41 = frac_41.reshape(new_shape)
        frac_32 = frac_32.reshape(new_shape)
        frac_42 = frac_42.reshape(new_shape)
        frac_c_41 = frac_c_41.reshape(new_shape)
        frac_c_42 = frac_c_42.reshape(new_shape)
        frac_c_43 = frac_c_43.reshape(new_shape)

    values[cond_a_mask] = _get_cross_section_values_a(
        property_values[cond_a_mask],
        p21[cond_a_mask],
        p31[cond_a_mask],
        p41[cond_a_mask],
        frac_21[cond_a_mask],
        frac_31[cond_a_mask],
        frac_41[cond_a_mask],
        average=average,
    )

    values[cond_b_mask] = _get_cross_section_values_b(
        property_values[cond_b_mask],
        p31[cond_b_mask],
        p41[cond_b_mask],
        p32[cond_b_mask],
        p42[cond_b_mask],
        frac_31[cond_b_mask],
        frac_41[cond_b_mask],
        frac_32[cond_b_mask],
        frac_42[cond_b_mask],
        average=average,
    )

    values[cond_c_mask] = _get_cross_section_values_c(
        property_values[cond_c_mask],
        p41[cond_c_mask],
        p42[cond_c_mask],
        p43[cond_c_mask],
        frac_c_41[cond_c_mask],
        frac_c_42[cond_c_mask],
        frac_c_43[cond_c_mask],
        average=average,
    )

    return values


def _get_cross_section_values_a(
    property_values, p21, p31, p41, frac_21, frac_31, frac_41, average=False
):
    # these are the property values at the intersection of the cross section section
    # with the tetrahedron edges
    interp_21 = p21 * frac_21 + property_values[:, 0]
    interp_31 = p31 * frac_31 + property_values[:, 0]
    interp_41 = p41 * frac_41 + property_values[:, 0]

    if average:
        return (interp_21 + interp_31 + interp_41) / 3

    else:
        return np.stack(
            [interp_21, interp_31, interp_41, np.zeros_like(interp_21)], axis=1
        )


def _get_cross_section_values_b(
    property_values,
    p31,
    p41,
    p32,
    p42,
    frac_31,
    frac_41,
    frac_32,
    frac_42,
    average=False,
):
    # these are the property values at the intersection of the cross section section
    # with the tetrahedron edges
    interp_31 = p31 * frac_31 + property_values[:, 0]
    interp_41 = p41 * frac_41 + property_values[:, 0]
    interp_32 = p32 * frac_32 + property_values[:, 1]
    interp_42 = p42 * frac_42 + property_values[:, 1]

    if average:
        return (interp_31 + interp_41 + interp_32 + interp_42) / 4

    else:
        # the order of these matters for calculating the cross sectional
        # area. the first and last points are at opposite ends of the quadrilateral
        return np.stack([interp_31, interp_32, interp_41, interp_42], axis=1)


def _get_cross_section_values_c(
    property_values, p41, p42, p43, frac_41, frac_42, frac_43, average=False
):
    # these are the property values at the intersection of the cross section
    # with the tetrahedron edges
    interp_41 = property_values[:, 3] - p41 * frac_41
    interp_42 = property_values[:, 3] - p42 * frac_42
    interp_43 = property_values[:, 3] - p43 * frac_43

    if average:
        return (interp_41 + interp_42 + interp_43) / 3

    else:
        return np.stack(
            [interp_41, interp_42, interp_43, np.zeros_like(interp_41)], axis=1
        )


def get_projected_intersections(intersections):
    # project the intersections coordinates onto a two dimensional plane

    # take two of the triangle/quadrilateral sides and calculate the unit vectors
    # the first of these vectors will be a basis vector
    axis_a = intersections[:, 1] - intersections[:, 0]
    axis_a /= np.linalg.norm(axis_a, axis=1)[:, None]

    other_side = intersections[:, 2] - intersections[:, 0]

    # find the vector normal to the plane of the triangle/quadrilateral, this will be a
    # basis vector, however all points will have the same magnitude in this direction
    axis_c = np.cross(axis_a, other_side)
    axis_c /= np.linalg.norm(axis_c, axis=1)[:, None]

    # find the final basis vector by finding the normal to the a and c basis vectors
    # no need to normalise as it is the cross product of two orthogonal unit vectors
    axis_b = np.cross(axis_a, axis_c)

    # define a transformation matrix to transform the points to our new basis
    basis = np.stack([axis_a, axis_b, axis_c], axis=2)
    transform = np.linalg.inv(basis)

    # finally transform the intersection coordinates. This syntax is equivalent to
    # np.dot(transform[0], intersections[0][0]) for all intersections in all triangles/
    # quadrilaterals.
    return np.einsum("ikj,ilj->ilk", transform, intersections), basis


def process_tetrahedra(tetrahedra, energies):
    all_tetrahedra = {}
    all_tetrahedra_energies = {}

    for spin, spin_energies in energies.items():
        data_shape = (len(spin_energies),) + tetrahedra.shape
        spin_tetrahedra = np.zeros(data_shape, dtype=int)
        spin_tetrahedra_energies = np.zeros(data_shape)

        for band_idx, band_energies in enumerate(spin_energies):
            band_tetrahedra_energies = band_energies[tetrahedra]

            sort_idx = np.argsort(band_tetrahedra_energies, axis=1)
            spin_tetrahedra_energies[band_idx, ...] = np.take_along_axis(
                band_tetrahedra_energies, sort_idx, axis=1
            )
            spin_tetrahedra[band_idx, ...] = np.take_along_axis(
                tetrahedra, sort_idx, axis=1
            )

        all_tetrahedra[spin] = spin_tetrahedra
        all_tetrahedra_energies[spin] = spin_tetrahedra_energies

    return all_tetrahedra, all_tetrahedra_energies


def get_tetrahedra_energy_diffs(tetrahedra_energies):
    e21 = {}
    e31 = {}
    e41 = {}
    e32 = {}
    e42 = {}
    e43 = {}

    for spin, s_tetrahedra_energies in tetrahedra_energies.items():
        # each energy difference has the shape nbands, ntetrahedra
        e21[spin] = s_tetrahedra_energies[:, :, 1] - s_tetrahedra_energies[:, :, 0]
        e31[spin] = s_tetrahedra_energies[:, :, 2] - s_tetrahedra_energies[:, :, 0]
        e41[spin] = s_tetrahedra_energies[:, :, 3] - s_tetrahedra_energies[:, :, 0]
        e32[spin] = s_tetrahedra_energies[:, :, 2] - s_tetrahedra_energies[:, :, 1]
        e42[spin] = s_tetrahedra_energies[:, :, 3] - s_tetrahedra_energies[:, :, 1]
        e43[spin] = s_tetrahedra_energies[:, :, 3] - s_tetrahedra_energies[:, :, 2]

    return e21, e31, e41, e32, e42, e43


def get_max_min_tetrahedra_energies(tetrahedra_energies):
    max_tetrahedra_energies = {}
    min_tetrahedra_energies = {}

    for spin, s_tetrahedra_energies in tetrahedra_energies.items():
        max_tetrahedra_energies[spin] = np.max(s_tetrahedra_energies, axis=2)
        min_tetrahedra_energies[spin] = np.min(s_tetrahedra_energies, axis=2)

    return max_tetrahedra_energies, min_tetrahedra_energies


def get_tetrahedra_cross_section_weights(
    reciprocal_lattice, kpoints, tetrahedra, e21, e31, e41
):
    # weight (b) defined by equation 3.4 in https://doi.org/10.1002/pssb.2220540211
    # this weight is not the Bloechl integrand but a scaling needed to obtain the
    # DOS directly from the tetrahedron cross section
    cross_section_weights = {}

    # volume is 6 * the volume of one tetrahedron
    volume = np.linalg.det(reciprocal_lattice) / len(kpoints)
    for spin, s_tetrahedra in tetrahedra.items():
        tetrahedra_kpoints = kpoints[s_tetrahedra]

        k1 = pbc_diff(tetrahedra_kpoints[:, :, 1], tetrahedra_kpoints[:, :, 0])
        k2 = pbc_diff(tetrahedra_kpoints[:, :, 2], tetrahedra_kpoints[:, :, 0])
        k3 = pbc_diff(tetrahedra_kpoints[:, :, 3], tetrahedra_kpoints[:, :, 0])

        k1_cart = np.dot(k1, reciprocal_lattice)
        k2_cart = np.dot(k2, reciprocal_lattice)
        k3_cart = np.dot(k3, reciprocal_lattice)

        contragradient = np.stack(
            [
                np.cross(k2_cart, k3_cart) / volume,
                np.cross(k3_cart, k1_cart) / volume,
                np.cross(k1_cart, k2_cart) / volume,
            ],
            axis=2,
        )

        energies = np.stack([e21[spin], e31[spin], e41[spin]], axis=2)
        b_vector = np.sum(contragradient * energies[..., None], axis=2)

        cross_section_weights[spin] = 1 / np.linalg.norm(b_vector, axis=2)

    return cross_section_weights



import numpy as np

from amset.interpolation.periodic import (
    PeriodicLinearInterpolator,
    group_bands_and_kpoints,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


class MRTACalculator(PeriodicLinearInterpolator):
    def get_mrta_factor(self, spin, band_a, kpoint_a, band_b, kpoint_b):
        bands, kpoints, single_factor = group_bands_and_kpoints(
            band_a, kpoint_a, band_b, kpoint_b
        )
        p = self.interpolate(spin, bands, kpoints)
        factor = 1 - np.dot(p[0], p[1:].T) / np.linalg.norm(p[0]) ** 2

        if single_factor:
            return factor[0]
        else:
            return factor



import logging
import os
import shutil
import sys
import tempfile
from pathlib import Path

import pytest
from monty.serialization import dumpfn, loadfn


@pytest.fixture(scope="session")
def test_dir():
    module_dir = Path(__file__).resolve().parent
    test_dir = module_dir / "test_data"
    return test_dir.resolve()


@pytest.fixture(scope="session")
def example_dir():
    root_dir = Path(__file__).resolve().parent.parent
    example_dir = root_dir / "examples"
    return example_dir.resolve()


@pytest.fixture
def log_to_stdout():
    # Set Logging
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch.setFormatter(formatter)
    root.addHandler(ch)


@pytest.fixture
def clean_dir():
    old_cwd = os.getcwd()
    newpath = tempfile.mkdtemp()
    os.chdir(newpath)
    yield
    os.chdir(old_cwd)
    shutil.rmtree(newpath)


@pytest.fixture(
    params=[
        "Si_227",
        "Fe_229",
        "S_58",
        "Rb2P3_69",
        "K2Au3_71",
        "LaI3_63",
        "KCeF4_123",
        "RbO2_129",
        "BaN2_15",
        "TiNi_11",
        "CaC2_2",
        "KNO3_160",
        "ZnO_186",
    ],
    ids=[
        "F cubic",
        "I cubic",
        "P orth",
        "I orth",
        "F orth",
        "C orth",
        "P tet",
        "I tet",
        "C mono",
        "P mono",
        "tri",
        "rhom",
        "hex",
    ],
    scope="session",
)
def symmetry_structure(test_dir, request):
    return loadfn(test_dir / "structures" / f"{request.param}.json.gz")


@pytest.fixture(scope="session")
def band_structure_data(test_dir):
    data = {}
    for band_structure_data_file in (test_dir / "band_structures").glob("*json*"):
        key = str(band_structure_data_file.name)
        for replace_str in [".json", ".gz", "band_structure_data"]:
            key = key.replace(replace_str, "")
        key = key.rstrip("_").lstrip("_")

        data[key] = loadfn(band_structure_data_file)
    return data


@pytest.fixture(scope="session")
def band_structures(band_structure_data):
    return {k: v["band_structure"] for k, v in band_structure_data.items()}


if __name__ == "__main__":
    # download test data
    materials = {
        "Si_227": "mp-149",  # F cubic
        "Fe_229": "mp-13",  # I cubic
        "S_58": "mp-558014",  # P orthorhombic
        "Rb2P3_69": "mp-2079",  # I orthorhombic
        "K2Au3_71": "mp-8700",  # F orthorhombic
        "LaI3_63": "mp-27979",  # C orthorhombic
        "KCeF4_123": "mp-1223451",  # P tetragonal
        "RbO2_129": "mp-12105",  # I tetragonal
        "BaN2_15": "mp-1001",  # C monoclinic
        "TiNi_11": "mp-1048",  # P monoclinic
        "CaC2_2": "mp-642822",  # triclinic
        "KNO3_160": "mp-6920",  # rhombohedral
        "ZnO_186": "mp-2133",  # hexagonal
    }
    from pymatgen.ext.matproj import MPRester

    _mpr = MPRester()
    _module_dir = Path(__file__).resolve().parent
    _structure_dir = _module_dir / "test_data" / "structures"

    if not _structure_dir.exists():
        _structure_dir.mkdir()

    for name, mp_id in materials.items():
        _s = _mpr.get_structure_by_material_id(mp_id)
        dumpfn(_s, _structure_dir / f"{name}.json.gz")



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple

import numpy as np

from amset.constants import nm_to_bohr, s_to_au
from amset.core.data import AmsetData

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class AbstractBasicScattering(ABC):

    name: str
    required_properties: Tuple[str]

    def __init__(self, properties, doping, temperatures, nbands):
        self.properties = properties
        self.doping = doping
        self.temperatures = temperatures
        self.nbands = nbands
        self.spins = list(nbands.keys())

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
        )

    @property
    @abstractmethod
    def rates(self):
        pass

    @classmethod
    def get_properties(cls, materials_properties):
        return {p: materials_properties[p] for p in cls.required_properties}

    @staticmethod
    def get_nbands(amset_data):
        return {s: len(amset_data.energies[s]) for s in amset_data.spins}


class ConstantRelaxationTime(AbstractBasicScattering):

    name = "CRT"
    required_properties = ("constant_relaxation_time",)

    def __init__(self, properties, doping, temperatures, nbands, shape):
        super().__init__(properties, doping, temperatures, nbands)
        rate = 1 / self.properties["constant_relaxation_time"]
        self._rates = {s: np.full(shape[s], rate) for s in self.spins}

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        shape = {
            s: amset_data.fermi_levels.shape + amset_data.energies[s].shape
            for s in amset_data.spins
        }
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
            shape,
        )

    @property
    def rates(self):
        # need to return rates with shape (nspins, ndops, ntemps, nbands, nkpoints)
        return self._rates


class MeanFreePathScattering(AbstractBasicScattering):

    name = "MFP"
    required_properties = ("mean_free_path",)

    def __init__(self, properties, doping, temperatures, nbands, rates):
        super().__init__(properties, doping, temperatures, nbands)
        self._rates = rates

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        rates = {}
        mfp = materials_properties["mean_free_path"] * nm_to_bohr
        ir_kpoints_idx = amset_data.ir_kpoints_idx
        for spin in amset_data.spins:
            vvelocities = amset_data.velocities_product[spin][..., ir_kpoints_idx]
            v = np.sqrt(np.diagonal(vvelocities, axis1=1, axis2=2))
            v = np.linalg.norm(v, axis=2)
            v[v < 0.005] = 0.005  # handle very small velocities
            velocities = np.tile(
                v, (len(amset_data.doping), len(amset_data.temperatures), 1, 1)
            )

            spin_rates = velocities * s_to_au / mfp
            rates[spin] = spin_rates[..., amset_data.ir_to_full_kpoint_mapping]
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
            rates,
        )

    @property
    def rates(self):
        # need to return rates with shape (nspins, ndops, ntemps, nbands, nkpoints)
        return self._rates


class ScaledRelaxationTime(AbstractBasicScattering):

    name = "SRT"
    required_properties = ("base_relaxation_time",)

    def __init__(self, properties, doping, temperatures, nbands, shape):
        super().__init__(properties, doping, temperatures, nbands)
        self._rates = {}
        for spin in self.spins:
            spin_rates = np.zeros(shape[spin])
            for i, temperature in enumerate(temperatures):
                lifetime = self.properties["base_relaxation_time"] * 300 / temperature
                spin_rates[:, i] = 1 / lifetime

            self._rates[spin] = spin_rates

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        shape = {
            s: amset_data.fermi_levels.shape + amset_data.energies[s].shape
            for s in amset_data.spins
        }
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
            shape,
        )

    @property
    def rates(self):
        # need to return rates with shape (nspins, ndops, ntemps, nbands, nkpoints)
        return self._rates



import abc
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

import numpy as np
from monty.serialization import loadfn
from pkg_resources import resource_filename

from amset import __version__
from amset.core.data import AmsetData
from amset.io import load_mesh
from amset.util import cast_dict_ndarray

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

amset_base_style = resource_filename("amset.plot", "amset_base.mplstyle")
revtex_style = resource_filename("amset.plot", "revtex.mplstyle")
base_total_color = "#193A4A"


class BaseMeshPlotter(abc.ABC):
    def __init__(self, mesh_data: Union[str, Path, AmsetData, dict]):
        if isinstance(mesh_data, (str, Path)):
            if "h5" not in str(mesh_data):
                raise ValueError(
                    "mesh.h5 file needed for plot. Run AMSET with write_mesh=True"
                )
            mesh_data = load_mesh(mesh_data)
        elif isinstance(mesh_data, AmsetData):
            mesh_data = mesh_data.to_dict(include_mesh=True)["mesh"]
        elif "mesh" in mesh_data:
            mesh_data = mesh_data["mesh"]
        elif not isinstance(mesh_data, dict):
            raise ValueError("Unrecognised data format")

        self._data = cast_dict_ndarray(mesh_data)
        self.spins = list(self.energies.keys())

    def __getattr__(self, item):
        return self._data[item]


class BaseTransportPlotter(abc.ABC):
    def __init__(self, data: Union[str, Path, AmsetData, dict]):
        if isinstance(data, (str, Path)):
            # guess the mesh filename
            data = loadfn(data)
        elif isinstance(data, AmsetData):
            data = data.to_dict(include_mesh=True)
        elif isinstance(data, dict):
            data = data
        else:
            raise ValueError("Unrecognised data format")

        self._data = cast_dict_ndarray(data)

    def __getattr__(self, item):
        return self._data[item]

    @property
    def has_mobility(self):
        return "mobility" in self._data and self._data["mobility"] is not None

    @property
    def mobility(self):
        return self.get_mobility()

    def get_mobility(self, mechanism="overall"):
        if not self.has_mobility:
            raise IndexError("Mobility does not exist in transport data")
        return self._data["mobility"][mechanism]

    @property
    def mechanisms(self):
        if not self.has_mobility:
            raise ValueError("Mobility must be in transport data to extract mechanisms")

        return tuple(self._data["mobility"].keys())


class BaseMultiTransportPlotter(abc.ABC):
    def __init__(self, data: List[Union[str, Path, AmsetData, dict, list]]):
        if not isinstance(data, (tuple, list)):
            data = [data]

        if isinstance(data, tuple):
            data = list(data)

        if len(data) < 2:
            raise ValueError("More than 2 transport data needed for plotting")

        new_data = []
        for d in data:
            if isinstance(d, (str, Path)):
                # guess the mesh filename
                d = loadfn(d)
            elif isinstance(d, AmsetData):
                d = d.to_dict(include_mesh=True)
            elif isinstance(d, dict):
                d = d
            else:
                raise ValueError("Unrecognised data format")

            new_data.append(cast_dict_ndarray(d))

        temperatures = set(new_data[0]["temperatures"])
        doping = set(new_data[0]["doping"])
        for d in new_data[1:]:
            if set(d["temperatures"]) != temperatures or set(d["doping"]) != doping:
                raise ValueError(
                    "Transport data contain inconsistent doping or temperatures"
                )

        self.temperatures = new_data[0]["temperatures"]
        self.doping = new_data[0]["doping"]
        self.n = len(new_data)
        self._data = new_data

    def __getattr__(self, item):
        return np.array([d[item] for d in self._data])

    @property
    def has_mobility(self):
        return all(["mobility" in d for d in self._data])

    @property
    def mobility(self):
        return self.get_mobility()

    def get_mobility(self, mechanism="overall"):
        if not self.has_mobility:
            raise IndexError("Mobility does not exist in all transport data")
        return np.array([d["mobility"][mechanism] for d in self._data])


@dataclass
class PlotData:
    """A class to represent basic plot data.

    Args:
        x_property: The x-axis property.
        y_property: The y-axis property.
        x_label: The x-axis label containing the unit.
        y_label: The y-axis label containing the unit.
        labels: The labels for each line on the plot. A list of strings of length m.
        x: The x-axis values as a numpy array with shape (n).
        y: The y-axis values with the shape (m, n).
    """

    x_property: str
    y_property: str
    x_label: str
    y_label: str
    labels: List[str]
    x: np.ndarray
    y: np.ndarray

    @property
    def comment(self) -> str:
        """Comment containing the labels, properties, and units."""
        comment = f"File generated using amset {__version__}"
        comment += f"\nx-axis: {self.x_label}"
        comment += f"\ny-axis: {self.y_label}"
        comment += f"\n{self.x_property.replace(' ', '_')}"

        for label in self.labels:
            comment += f"  {label.replace(' ', '')}"

        return comment

    @property
    def data(self) -> np.ndarray:
        """Data (containing x and y values) as a numpy array."""
        return np.vstack([self.x, self.y])

    def to_file(self, filename: [str, Path], fmt: str = "%.8e"):
        """Write plot data to file.

        Args:
            filename: The file name.
            fmt: Number format specifier.
        """
        np.savetxt(filename, self.data.T, fmt=fmt, header=self.comment)


def write_plot_data(
    plot_data: List[PlotData],
    prefix: Optional[Union[str, Path]] = None,
    directory: Optional[Union[str, Path]] = None,
) -> List[Path]:
    """Write plot data to files.

    Args:
        plot_data: A list of PlotData objects. The y_property attribute is used as the
            filename.
        prefix: A prefix for the data files.
        directory: The directory in which to store the data files.

    Returns:
        A list of path objects for each file.
    """
    if directory is None:
        directory = ""
    directory = Path(directory)

    if prefix is None:
        prefix = ""
    prefix = str(prefix)
    if len(prefix) > 0 and prefix[-1] != "_":
        prefix += "_"

    filenames = []
    for data in plot_data:
        propname = data.y_property.replace(" ", "_")
        filename = directory / f"{prefix}{propname}.dat"
        data.to_file(filename)
        filenames.append(filename)

    return filenames



"""
Module defining io functions.
"""

from typing import Any, Dict

import numpy as np
from monty.serialization import dumpfn, loadfn

from amset.util import cast_dict_list, logger, validate_settings

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


def write_settings(settings: Dict[str, Any], filename: str):
    """Write amset configuration settings to a formatted yaml file.

    Args:
        settings: The configuration settings.
        filename: A filename.
    """
    settings = cast_dict_list(settings)
    dumpfn(settings, filename, indent=4, default_flow_style=False)


def load_settings(filename: str) -> Dict[str, Any]:
    """Load amset configuration settings from a yaml file.

    If the settings file does not contain a required parameter, the default
    value will be added to the configuration.

    An example file is given in *amset/examples/example_settings.yaml*.

    Args:
        filename: Path to settings file.

    Returns:
        The settings, with any missing values set according to the amset defaults.
    """
    logger.info(f"Loading settings from: {filename}")
    settings = loadfn(filename)

    return validate_settings(settings)


def write_mesh(mesh_data, filename="mesh.h5"):
    import h5py
    from pymatgen.core.structure import Structure

    with h5py.File(filename, "w") as f:

        def add_data(name, data):
            if isinstance(data, np.ndarray):
                f.create_dataset(name, data=data, compression="gzip")
            elif isinstance(data, Structure):
                f["structure"] = np.string_(data.to_json())
            elif isinstance(data, (tuple, list)):
                data = np.array(data)
                if isinstance(data[0], str):
                    data = data.astype("S")
                f.create_dataset(name, data=data)
            elif data is None:
                f.create_dataset(name, data=False)
            else:
                f.create_dataset(name, data=data)

        for key, value in mesh_data.items():
            if isinstance(value, dict):
                # dict entries are given for different spins
                for spin, spin_value in value.items():
                    key = f"{key}_{spin.name}"
                    add_data(key, spin_value)
            else:
                add_data(key, value)


def load_mesh(filename):
    import h5py
    from pymatgen.core.structure import Structure

    from amset.constants import str_to_spin

    def read_data(name, data):
        if name == "structure":
            data_str = np.string_(data[()]).decode()
            return Structure.from_str(data_str, fmt="json")
        if name == "scattering_labels":
            return data[()].astype("U13")  # decode string
        if name == "vb_idx":
            d = data[()]
            return d if d is not False else None
        return data[()]

    mesh_data = {}
    with h5py.File(filename, "r") as f:
        for key, value in f.items():
            if "_up" in key or "_down" in key:
                spin = str_to_spin[key.split("_")[-1]]
                key = key.replace(f"_{spin.name}", "")
                if key not in mesh_data:
                    mesh_data[key] = {}
                mesh_data[key][spin] = read_data(key, value)
            else:
                mesh_data[key] = read_data(key, value)

    return mesh_data



import warnings
from pathlib import Path

import numpy as np
from tqdm.auto import tqdm

from amset.constants import int_to_spin, numeric_types, spin_to_int
from amset.wavefunction.common import (
    get_gpoint_indices,
    get_gpoints,
    get_min_gpoints,
    get_overlap,
    is_ncl,
    sample_random_kpoints,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


def get_wavefunction(wavecar="WAVECAR", vasp_type=None, directory=None):
    from pymatgen.io.vasp import Wavecar

    if directory:
        wavecar = Path(directory) / wavecar

    return Wavecar(wavecar, vasp_type=vasp_type)


def get_wavefunction_coefficients(
    wavefunction, iband=None, ikpoints=None, encut=500, pbar=True
):
    if encut > wavefunction.encut:
        warnings.warn(
            "Selected encut greater than encut of calculation. Many coefficients will "
            "likely be zero."
        )

    valid_gpoints = get_gpoints(wavefunction.b, wavefunction._nbmax, encut)
    min_gpoint, num_gpoint = get_min_gpoints(wavefunction._nbmax)
    valid_indices = get_gpoint_indices(valid_gpoints, min_gpoint, num_gpoint)

    if not iband:
        iband = {}

    coeffs = {}
    for spin_idx in range(wavefunction.spin):
        spin = int_to_spin[spin_idx]
        spin_iband = iband.get(spin, None)

        coeffs[spin] = _get_spin_wavefunction_coefficients(
            wavefunction,
            spin,
            valid_indices,
            min_gpoint,
            num_gpoint,
            iband=spin_iband,
            ikpoints=ikpoints,
            pbar=pbar,
        )
    return coeffs, valid_gpoints


def _get_spin_wavefunction_coefficients(
    wavefunction,
    spin,
    valid_indices,
    min_gpoint,
    num_gpoint,
    iband=None,
    ikpoints=None,
    pbar=True,
):
    from amset.constants import output_width

    set_valid_indices = set(valid_indices)

    # mapping from each valid gpoint index to final gpoint array
    indices_map = np.full(np.max(valid_indices) + 1, -1, dtype=int)
    indices_map[valid_indices] = np.arange(len(valid_indices))

    ispin = spin_to_int[spin]
    if wavefunction.spin == 1:
        original_coeffs = wavefunction.coeffs
    else:
        original_coeffs = wavefunction.coeffs[ispin]

    ncl = "ncl" == wavefunction.vasp_type

    if iband is None:
        # take all bands
        iband = np.arange(len(original_coeffs[0]), dtype=int)

    if ikpoints is None:
        ikpoints = np.arange(len(wavefunction.kpoints))

    elif isinstance(iband, numeric_types):
        iband = [iband]

    ncoeffs = len(valid_indices)
    nkpoints = len(ikpoints)
    if ncl:
        coeffs = np.zeros((len(iband), nkpoints, ncoeffs, 2), dtype=complex)
    else:
        coeffs = np.zeros((len(iband), nkpoints, ncoeffs), dtype=complex)

    if pbar:
        ikpoints = tqdm(ikpoints, ncols=output_width)

    for j, nk in enumerate(ikpoints):
        gpoints_to_keep, indices_to_keep = _get_valid_state_coefficients(
            wavefunction, nk, min_gpoint, num_gpoint, set_valid_indices
        )
        coeff_indices = indices_map[indices_to_keep]

        if np.any(coeff_indices == -1):
            raise RuntimeError("Something went wrong mapping coefficients")

        for i, nb in enumerate(iband):
            coeffs[i, j, coeff_indices] = original_coeffs[nk][nb].T[gpoints_to_keep]

    if ncl:
        coeffs /= np.linalg.norm(coeffs, axis=(2, 3))[..., None, None]
    else:
        coeffs /= np.linalg.norm(coeffs, axis=2)[..., None]

    return coeffs


def _get_valid_state_coefficients(
    wavefunction, nkpoint, min_gpoint, num_gpoint, set_valid_indices
):
    """
    Returns the index of the gpoints in the Gpoints array and the correspoding
    'index' identifier.
    """
    gpoints = wavefunction.Gpoints[nkpoint]
    indices = get_gpoint_indices(gpoints, min_gpoint, num_gpoint)

    # map the indices to the original gpoint array
    indices_map = np.full(np.max(indices) + 1, -1, dtype=int)
    indices_map[indices] = np.arange(len(indices))

    # find which indices are in the list of valid indices
    set_indices = set(indices)
    indices_to_keep = set_valid_indices.intersection(set_indices)
    indices_to_keep = np.array(list(indices_to_keep))

    to_keep = indices_map[indices_to_keep]
    return to_keep, indices_to_keep


def get_converged_encut(
    wavefunction,
    iband=None,
    ikpoints=None,
    max_encut=None,
    n_samples=1000,
    std_tol=0.002,
):
    nspins = wavefunction.spin

    if max_encut is None:
        max_encut = wavefunction.encut

    if ikpoints is None:
        ikpoints = np.arange(len(wavefunction.kpoints))

    if iband is None:
        iband = {}
        for ispin in range(wavefunction.spin):
            spin = int_to_spin[ispin]
            if wavefunction.spin == 1:
                iband[spin] = np.arange(len(wavefunction.coeffs[0]))
            else:
                iband[spin] = np.arange(len(wavefunction.coeffs[ispin][0]))

    # this is a little different to usual iband
    sample_iband = {s: len(b) for s, b in iband.items()}
    coeffs, _ = get_wavefunction_coefficients(
        wavefunction, encut=max_encut, pbar=False, iband=iband, ikpoints=ikpoints
    )

    sample_points = sample_random_kpoints(
        nspins, np.arange(len(ikpoints)), sample_iband, n_samples
    )
    origin = sample_points[0]
    sample_points = sample_points[1:]

    true_overlaps = get_overlaps(coeffs, origin, sample_points)

    # filter points to only include these with reasonable overlaps
    mask = true_overlaps > 0.05
    true_overlaps = true_overlaps[mask]
    sample_points = sample_points[mask]

    for encut in np.arange(200, max_encut, 50):
        coeffs, _ = get_wavefunction_coefficients(
            wavefunction, encut=encut, pbar=False, iband=iband, ikpoints=ikpoints
        )
        fake_overlaps = get_overlaps(coeffs, origin, sample_points)

        diff = (true_overlaps / fake_overlaps) - 1
        if diff.std() < std_tol:
            return encut

    return max_encut


def get_overlaps(coeffs, origin, points):
    ncoeffs = list(coeffs.values())[0].shape[2]
    ncl = is_ncl(coeffs)
    if ncl:
        select_coeffs = np.zeros((len(points), ncoeffs, 2), dtype=complex)
    else:
        select_coeffs = np.zeros((len(points), ncoeffs), dtype=complex)

    origin_spin = int_to_spin[origin[0]]
    origin = coeffs[origin_spin][origin[1], origin[2]]

    for i, (s_idx, b_idx, k_idx) in enumerate(points):
        spin = int_to_spin[s_idx]
        select_coeffs[i] = coeffs[spin][b_idx, k_idx]

    return get_overlap(origin, select_coeffs)



"""
Main integration test for amset calculations

Tests running amset from:
- A vasprun.xml file
- A directory with override settings specified

The following tests are performed:
- don't write mesh, using projections + deformation potential tuple + single elastic
  constant/piezoelectric for Silicon
- write mesh file, using projections + deformation potential tuple + single elastic
  constant/piezoelectric for Silicon
- use wavefunction coefficients + using deformation potential file + full elastic
  constant/piezoelectric for Silicon
- use wavefunction coefficients + using deformation potential file + full elastic
  constant/piezoelectric + no cache for Silicon
- use wavefunction coefficients + using deformation potential file + full elastic
  constant/piezoelectric for Gallium Arsenide
- don't write mesh, using projections + deformation potential tuple + single elastic
  constant/piezoelectric for K2ReF6 (tricky spin polarized system)
"""
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict

import numpy as np
import pytest
from monty.serialization import dumpfn

from amset.core.run import Runner

si_settings_no_mesh: Dict[str, Any] = {
    "interpolation_factor": 5,
    "doping": [-1e15, -1e16, -1e17],
    "deformation_potential": (6.5, 6.5),
    "elastic_constant": 190,
    "static_dielectric": 13.1,
    "use_projections": True,
    "nworkers": 1,
}
si_settings_mesh = deepcopy(si_settings_no_mesh)
si_settings_mesh.update({"write_mesh": True})
si_settings_wavefunction = deepcopy(si_settings_no_mesh)
si_settings_wavefunction.update(
    {
        "deformation_potential": "deformation.h5",
        "elastic_constant": [
            [144, 53, 53, 0, 0, 0],
            [53, 144, 53, 0, 0, 0],
            [53, 53, 144, 0, 0, 0],
            [0, 0, 0, 75, 0, 0],
            [0, 0, 0, 0, 75, 0],
            [0, 0, 0, 0, 0, 75],
        ],
        "static_dielectric": [[11.7, 0, 0], [0, 11.7, 0], [0, 0, 11.7]],
        "use_projections": False,
    }
)
si_settings_wavefunction_nocache = deepcopy(si_settings_wavefunction)
si_settings_wavefunction_nocache.update({"cache_wavefunction": False})
si_transport_projections = {
    ("mobility", ("overall", (0, 0))): 1576.074528381305,
    ("mobility", ("overall", (-1, 0))): 817.7095423715324,
    ("seebeck", (0, 0)): -1008.6448354488775,
    ("seebeck", (-1, 0)): -710.4420172707946,
    ("conductivity", (0, 0)): 26.524979568158074,
    ("conductivity", (-1, 0)): 1310.121658393053,
    ("electronic_thermal_conductivity", (0, 0)): 0.0013629136608729912,
}
si_transport_wavefunction = {
    ("mobility", ("overall", (0, 0))): 1234.1461746826737,
    ("mobility", ("overall", (-1, 0))): 667.3798557475134,
    ("seebeck", (0, 0)): -960.4571567090115,
    ("seebeck", (-1, 0)): -711.2202530136929,
    ("conductivity", (0, 0)): 21.233368948315626,
    ("conductivity", (-1, 0)): 1069.265903429082,
    ("electronic_thermal_conductivity", (0, 0)): 0.0018031827294972345,
}

gaas_settings_wavefunction = {
    "doping": [-3e13],
    "temperatures": [201, 290, 401, 506, 605, 789, 994],
    "bandgap": 1.33,
    "interpolation_factor": 5,
    "deformation_potential": (1.2, 8.6),
    "elastic_constant": 139.7,
    "donor_charge": 1,
    "acceptor_charge": 1,
    "static_dielectric": 12.18,
    "high_frequency_dielectric": 10.32,
    "pop_frequency": 8.16,
    "nworkers": 1,
}
gaas_transport = {
    ("mobility", ("overall", (0, 0))): 21305.275888767374,
    ("mobility", ("overall", (0, -1))): 2428.11416884464,
    ("seebeck", (0, 0)): -858.4131242931522,
    ("seebeck", (0, -1)): -608.3181725103633,
    ("conductivity", (0, 0)): 10.24043666456572,
    ("conductivity", (0, -1)): 241.9083618988777,
    ("electronic_thermal_conductivity", (0, -1)): 0.026674584564867737,
}

test_data = [
    pytest.param(
        "Si",
        si_settings_no_mesh,
        si_transport_projections,
        0.001,
        ["transport", "!mesh"],
        ["ADP", "IMP"],
        id="Si (no mesh, projections, simple scats)",
    ),
    pytest.param(
        "Si",
        si_settings_mesh,
        si_transport_projections,
        0.001,
        ["transport", "mesh"],
        ["ADP", "IMP"],
        id="Si (mesh, projections, simple scats)",
    ),
    pytest.param(
        "Si",
        si_settings_wavefunction,
        si_transport_wavefunction,
        0.001,
        ["transport", "!mesh"],
        ["ADP", "IMP"],
        id="Si (wavefunction, best scats)",
    ),
    pytest.param(
        "Si",
        si_settings_wavefunction_nocache,
        si_transport_wavefunction,
        0.001,
        ["transport", "!mesh"],
        ["ADP", "IMP"],
        id="Si (wavefunction, best scats, no cache)",
    ),
    pytest.param(
        "GaAs",
        gaas_settings_wavefunction,
        gaas_transport,
        0.001,
        ["transport", "!mesh"],
        ["ADP", "IMP", "POP"],
        id="GaAs (wavefunction, best scats)",
    ),
]


@pytest.mark.usefixtures("clean_dir")
@pytest.mark.parametrize("system,settings,transport,max_aniso,files,scats", test_data)
def test_run_amset_from_vasprun(
    example_dir, system, settings, transport, max_aniso, files, scats
):
    vasprun, settings = _prep_inputs(example_dir, system, settings)
    runner = Runner.from_vasprun(vasprun, settings)
    amset_data = runner.run()
    _validate_data(amset_data, transport, max_aniso, files, scats)


@pytest.mark.usefixtures("clean_dir")
@pytest.mark.parametrize(
    "system,settings,transport,max_aniso,files,scats", [test_data[0]]
)
def test_run_amset_from_directory(
    example_dir, system, settings, transport, max_aniso, files, scats
):
    vasprun, settings = _prep_inputs(example_dir, system, settings)
    dumpfn(settings, "settings.yaml")
    runner = Runner.from_directory(".", input_file=vasprun, settings_override=settings)
    amset_data = runner.run()
    _validate_data(amset_data, transport, max_aniso, files, scats)


@pytest.mark.usefixtures("clean_dir")
def test_run_tricky_spin_polarized(band_structure_data):
    settings = {
        "interpolation_factor": 2,
        "temperatures": [300],
        "doping": [1e15],
        "deformation_potential": (6.5, 6.5),
        "elastic_constant": 190,
        "use_projections": True,
        "scattering_type": ["ADP"],
        "nworkers": 1,
    }
    files = ["transport", "!mesh"]
    scats = ["ADP"]
    transport = {
        ("mobility", ("overall", (0, 0))): 1.261213643946453,
        ("seebeck", (0, 0)): 1356.1341462605026,
    }

    bs = band_structure_data["tricky_sp"]["band_structure"]
    nelect = band_structure_data["tricky_sp"]["nelect"]
    runner = Runner(bs, nelect, settings)
    amset_data = runner.run()
    _validate_data(amset_data, transport, 1, files, scats)


def _prep_inputs(example_dir, system, settings):
    """Prepare calculation inputs"""

    input_dir = example_dir / system
    vasprun = input_dir / "vasprun.xml.gz"
    settings["wavefunction_coefficients"] = input_dir / "wavefunction.h5"

    ed = settings.get("deformation_potential")
    if isinstance(ed, str):
        settings["deformation_potential"] = input_dir / ed
    return vasprun, settings


def _validate_data(amset_data, transport, max_aniso, files, scats):
    """Validate calculation outputs"""

    # assert results are isotropic (often a warning sign that something is wrong)
    for temp_mobility in amset_data.mobility["overall"][:, 0]:
        eigs = np.linalg.eigvals(temp_mobility)
        assert np.abs(1 - (np.max(eigs) / np.min(eigs))) < max_aniso

    # check correct files are written out
    ls = [p.as_posix() for p in Path(".").glob("*")]
    for file in files:
        if file[0] == "!":
            assert all([file[1:] not in x for x in ls])
        else:
            assert any([file in x for x in ls])

    # check transport properties
    for (prop, loc), expected in transport.items():
        calculated = getattr(amset_data, prop)
        if prop == "mobility":
            value = calculated[loc[0]][loc[1]]
        else:
            value = calculated[loc]

        assert value.shape == (3, 3)

        value = np.average(np.linalg.eigvalsh(value))
        print("('{}', {}): {},".format(prop, loc, value))

        # assert values agree to within 1 %
        assert (
            np.abs(1 - value / expected) < 0.01
        ), f"property: {prop}, loc: {loc}, differs by more than 1%: calculated: {value}, expected: {expected}"

    # check scattering types
    assert set(amset_data.scattering_labels) == set(scats)
